"""Enhanced semantic validation for YARA AST."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from yaraast.types.module_loader import ModuleLoader
from yaraast.types.type_system import FunctionDefinition, TypeChecker, TypeEnvironment
from yaraast.visitor import ASTVisitor

if TYPE_CHECKING:
    from yaraast.ast.base import Location, YaraFile
    from yaraast.ast.expressions import Expression, FunctionCall
    from yaraast.ast.rules import Rule
    from yaraast.ast.strings import StringDefinition


@dataclass
class ValidationError:
    """Rich validation error with location information."""

    message: str
    location: Location | None = None
    error_type: str = "semantic"
    severity: str = "error"  # "error", "warning", "info"
    suggestion: str | None = None

    def __str__(self) -> str:
        """Format error message with location."""
        if self.location:
            return f"{self.location.file}:{self.location.line}:{self.location.column}: {self.severity}: {self.message}"
        return f"{self.severity}: {self.message}"

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization."""
        result = {
            "message": self.message,
            "error_type": self.error_type,
            "severity": self.severity,
        }
        if self.location:
            result["location"] = {
                "file": self.location.file,
                "line": self.location.line,
                "column": self.location.column,
            }
        if self.suggestion:
            result["suggestion"] = self.suggestion
        return result


@dataclass
class ValidationResult:
    """Result of semantic validation."""

    is_valid: bool = True
    errors: list[ValidationError] = field(default_factory=list)
    warnings: list[ValidationError] = field(default_factory=list)

    def add_error(
        self,
        message: str,
        location: Location | None = None,
        suggestion: str | None = None,
    ) -> None:
        """Add validation error."""
        error = ValidationError(message, location, "semantic", "error", suggestion)
        self.errors.append(error)
        self.is_valid = False

    def add_warning(
        self,
        message: str,
        location: Location | None = None,
        suggestion: str | None = None,
    ) -> None:
        """Add validation warning."""
        warning = ValidationError(message, location, "semantic", "warning", suggestion)
        self.warnings.append(warning)

    def combine(self, other: ValidationResult) -> None:
        """Combine with another validation result."""
        self.errors.extend(other.errors)
        self.warnings.extend(other.warnings)
        if not other.is_valid:
            self.is_valid = False

    @property
    def total_issues(self) -> int:
        """Total number of issues (errors + warnings)."""
        return len(self.errors) + len(self.warnings)


class StringIdentifierValidator(ASTVisitor[None]):
    """Validator for string identifier uniqueness within rules."""

    def __init__(self, result: ValidationResult) -> None:
        self.result = result
        self.current_rule_strings: set[str] = set()
        self.current_rule_name: str | None = None

    def visit_rule(self, node: Rule) -> None:
        """Validate string identifier uniqueness within rule."""
        self.current_rule_strings.clear()
        self.current_rule_name = node.name

        # Check all string definitions in the rule
        for string_def in node.strings:
            self.visit(string_def)

    def visit_string_definition(self, node: StringDefinition) -> None:
        """Check string identifier uniqueness."""
        # Remove $ prefix if present for comparison
        identifier = node.identifier
        identifier = identifier.removeprefix("$")

        # Anonymous strings (just $) are allowed to be duplicated in YARA
        if identifier == "":
            # Anonymous string - these can be duplicated
            return

        if identifier in self.current_rule_strings:
            self.result.add_error(
                f"Duplicate string identifier '${identifier}' in rule '{self.current_rule_name}'",
                node.location,
                f"String identifiers must be unique within each rule. Consider renaming to '${identifier}_2' or similar.",
            )
        else:
            self.current_rule_strings.add(identifier)

    # Default implementations for other node types
    def visit_yara_file(self, node) -> None:
        """Visit YARA file node - handled by rule-specific validation."""

    def visit_import(self, node) -> None:
        """Visit import node - imports don't affect string identifier validation."""

    def visit_include(self, node) -> None:
        """Visit include node - includes don't affect string identifier validation."""

    def visit_tag(self, node) -> None:
        """Visit tag node - tags don't affect string identifier validation."""

    def visit_plain_string(self, node) -> None:
        self.visit_string_definition(node)

    def visit_hex_string(self, node) -> None:
        self.visit_string_definition(node)

    def visit_regex_string(self, node) -> None:
        self.visit_string_definition(node)

    def visit_string_modifier(self, node) -> None:
        """Visit string modifier node - modifiers don't affect identifier uniqueness."""

    def visit_hex_token(self, node) -> None:
        """Visit hex token node - hex tokens don't affect identifier validation."""

    def visit_hex_byte(self, node) -> None:
        """Visit hex byte node - hex bytes don't affect identifier validation."""

    def visit_hex_wildcard(self, node) -> None:
        """Visit hex wildcard node - wildcards don't affect identifier validation."""

    def visit_hex_jump(self, node) -> None:
        """Visit hex jump node - jumps don't affect identifier validation."""

    def visit_hex_alternative(self, node) -> None:
        """Visit hex alternative node - alternatives don't affect identifier validation."""

    def visit_hex_nibble(self, node) -> None:
        """Visit hex nibble node - nibbles don't affect identifier validation."""

    def visit_expression(self, node) -> None:
        """Visit expression node - expressions don't affect string identifier validation."""

    def visit_identifier(self, node) -> None:
        """Visit identifier node - identifiers don't affect string identifier validation."""

    def visit_string_identifier(self, node) -> None:
        """Visit string identifier node - string identifiers don't affect definition validation."""

    def visit_string_wildcard(self, node) -> None:
        """Visit string wildcard node - string wildcards don't affect definition validation."""

    def visit_string_count(self, node) -> None:
        """Visit string count node - string counts don't affect identifier validation."""

    def visit_string_offset(self, node) -> None:
        """Visit string offset node - string offsets don't affect identifier validation."""

    def visit_string_length(self, node) -> None:
        """Visit string length node - string lengths don't affect identifier validation."""

    def visit_integer_literal(self, node) -> None:
        """Visit integer literal node - integer literals don't affect identifier validation."""

    def visit_double_literal(self, node) -> None:
        """Visit double literal node - double literals don't affect identifier validation."""

    def visit_string_literal(self, node) -> None:
        """Visit string literal node - string literals don't affect identifier validation."""

    def visit_boolean_literal(self, node) -> None:
        """Visit boolean literal node - boolean literals don't affect identifier validation."""

    def visit_binary_expression(self, node) -> None:
        """Visit binary expression node - binary expressions don't affect identifier validation."""

    def visit_unary_expression(self, node) -> None:
        """Visit unary expression node - unary expressions don't affect identifier validation."""

    def visit_parentheses_expression(self, node) -> None:
        """Visit parentheses expression node - parentheses don't affect identifier validation."""

    def visit_set_expression(self, node) -> None:
        """Visit set expression node - set expressions don't affect identifier validation."""

    def visit_range_expression(self, node) -> None:
        """Visit range expression node - range expressions don't affect identifier validation."""

    def visit_function_call(self, node) -> None:
        """Visit function call node - function calls don't affect identifier validation."""

    def visit_array_access(self, node) -> None:
        """Visit array access node - array access doesn't affect identifier validation."""

    def visit_member_access(self, node) -> None:
        """Visit member access node - member access doesn't affect identifier validation."""

    def visit_condition(self, node) -> None:
        """Visit condition node - conditions don't affect identifier validation."""

    def visit_for_expression(self, node) -> None:
        """Visit for expression node - for expressions don't affect identifier validation."""

    def visit_for_of_expression(self, node) -> None:
        """Visit for-of expression node - for-of expressions don't affect identifier validation."""

    def visit_at_expression(self, node) -> None:
        """Visit at expression node - at expressions don't affect identifier validation."""

    def visit_in_expression(self, node) -> None:
        """Visit in expression node - in expressions don't affect identifier validation."""

    def visit_of_expression(self, node) -> None:
        """Visit of expression node - of expressions don't affect identifier validation."""

    def visit_meta(self, node) -> None:
        """Visit meta node - meta fields don't affect identifier validation."""

    def visit_module_reference(self, node) -> None:
        """Visit module reference node - module references don't affect identifier validation."""

    def visit_dictionary_access(self, node) -> None:
        """Visit dictionary access node - dictionary access doesn't affect identifier validation."""

    def visit_comment(self, node) -> None:
        """Visit comment node - comments don't affect identifier validation."""

    def visit_comment_group(self, node) -> None:
        """Visit comment group node - comment groups don't affect identifier validation."""

    def visit_defined_expression(self, node) -> None:
        """Visit defined expression node - defined expressions don't affect identifier validation."""

    def visit_regex_literal(self, node) -> None:
        """Visit regex literal node - regex literals don't affect identifier validation."""

    def visit_string_operator_expression(self, node) -> None:
        """Visit string operator expression node - string operators don't affect identifier validation."""

    # Add missing abstract methods
    def visit_extern_import(self, node) -> None:
        """Visit extern import node - extern imports don't affect identifier validation."""

    def visit_extern_namespace(self, node) -> None:
        """Visit extern namespace node - extern namespaces don't affect identifier validation."""

    def visit_extern_rule(self, node) -> None:
        """Visit extern rule node - extern rules don't affect identifier validation."""

    def visit_extern_rule_reference(self, node) -> None:
        """Visit extern rule reference node - extern rule references don't affect identifier validation."""

    def visit_in_rule_pragma(self, node) -> None:
        """Visit in-rule pragma node - pragmas don't affect identifier validation."""

    def visit_pragma(self, node) -> None:
        """Visit pragma node - pragmas don't affect identifier validation."""

    def visit_pragma_block(self, node) -> None:
        """Visit pragma block node - pragma blocks don't affect identifier validation."""


class FunctionCallValidator(ASTVisitor[None]):
    """Validator for function calls and module function existence."""

    def __init__(self, result: ValidationResult, env: TypeEnvironment) -> None:
        self.result = result
        self.env = env
        self.module_loader = ModuleLoader()

    def visit_function_call(self, node: FunctionCall) -> None:
        """Validate function call existence and arity."""
        # Check if it's a module function call (e.g., pe.imphash)
        if "." in node.function:
            parts = node.function.split(".", 1)
            if len(parts) == 2:
                module_name, func_name = parts
                self._validate_module_function_call(node, module_name, func_name)
        else:
            # Built-in function call
            self._validate_builtin_function_call(node)

        # Recursively validate function arguments
        for arg in node.arguments:
            self.visit(arg)

    def _validate_module_function_call(
        self,
        node: FunctionCall,
        module_name: str,
        func_name: str,
    ) -> None:
        """Validate module function call."""
        # Check if module is imported
        if not self.env.has_module(module_name):
            self.result.add_error(
                f"Module '{module_name}' not imported, cannot call '{node.function}'",
                node.location,
                f"Add 'import \"{module_name}\"' at the top of your file.",
            )
            return

        # Get actual module name (handles aliases)
        actual_module = self.env.get_module_name(module_name)
        if not actual_module:
            return

        # Get module definition
        module_def = self.module_loader.get_module(actual_module)
        if not module_def:
            self.result.add_warning(
                f"Module definition for '{actual_module}' not found, cannot validate function '{func_name}'",
                node.location,
            )
            return

        # Check if function exists in module
        if func_name not in module_def.functions:
            available_funcs = list(module_def.functions.keys())
            self.result.add_error(
                f"Function '{func_name}' not found in module '{actual_module}'",
                node.location,
                (
                    f"Available functions: {', '.join(available_funcs)}"
                    if available_funcs
                    else "No functions available"
                ),
            )
            return

        # Validate function arity and parameter types
        func_def = module_def.functions[func_name]
        self._validate_function_arity(node, func_def)

    def _validate_builtin_function_call(self, node: FunctionCall) -> None:
        """Validate built-in function call."""
        builtin_functions = {
            # Integer reading functions
            "uint8": (1, 1),  # (min_args, max_args)
            "uint16": (1, 1),
            "uint32": (1, 1),
            "int8": (1, 1),
            "int16": (1, 1),
            "int32": (1, 1),
            # Big-endian variants
            "uint8be": (1, 1),
            "uint16be": (1, 1),
            "uint32be": (1, 1),
            "int8be": (1, 1),
            "int16be": (1, 1),
            "int32be": (1, 1),
            # Little-endian variants
            "uint16le": (1, 1),
            "uint32le": (1, 1),
            "int16le": (1, 1),
            "int32le": (1, 1),
        }

        func_name = node.function
        if func_name in builtin_functions:
            min_args, max_args = builtin_functions[func_name]
            actual_args = len(node.arguments)

            if actual_args < min_args:
                self.result.add_error(
                    f"Function '{func_name}' expects at least {min_args} argument(s), got {actual_args}",
                    node.location,
                )
            elif actual_args > max_args:
                self.result.add_error(
                    f"Function '{func_name}' expects at most {max_args} argument(s), got {actual_args}",
                    node.location,
                )
        else:
            # Unknown function - could be user-defined or module function without prefix
            self.result.add_warning(
                f"Unknown function '{func_name}'. If this is a module function, use 'module.{func_name}' syntax.",
                node.location,
            )

    def _validate_function_arity(
        self,
        node: FunctionCall,
        func_def: FunctionDefinition,
    ) -> None:
        """Validate function argument count and types."""
        expected_args = len(func_def.parameters)
        actual_args = len(node.arguments)

        if actual_args != expected_args:
            param_names = [p[0] for p in func_def.parameters]
            self.result.add_error(
                f"Function '{func_def.name}' expects {expected_args} argument(s) ({', '.join(param_names)}), got {actual_args}",
                node.location,
            )
            return

        # Parameter type validation is handled by the type system
        # Here we only validate function existence and arity

    # Recursive visitation methods
    def visit_binary_expression(self, node) -> None:
        self.visit(node.left)
        self.visit(node.right)

    def visit_unary_expression(self, node) -> None:
        self.visit(node.operand)

    def visit_member_access(self, node) -> None:
        # Don't validate module.function as function call here
        # It's handled in function_call validation
        self.visit(node.object)

    def visit_parentheses_expression(self, node) -> None:
        self.visit(node.expression)

    def visit_set_expression(self, node) -> None:
        for elem in node.elements:
            self.visit(elem)

    def visit_range_expression(self, node) -> None:
        self.visit(node.low)
        self.visit(node.high)

    def visit_array_access(self, node) -> None:
        self.visit(node.array)
        self.visit(node.index)

    def visit_for_expression(self, node) -> None:
        self.visit(node.iterable)
        self.visit(node.body)

    def visit_for_of_expression(self, node) -> None:
        self.visit(node.string_set)
        if node.condition:
            self.visit(node.condition)

    def visit_at_expression(self, node) -> None:
        self.visit(node.offset)

    def visit_in_expression(self, node) -> None:
        self.visit(node.range)

    def visit_of_expression(self, node) -> None:
        self.visit(node.quantifier)
        self.visit(node.string_set)

    # Default implementations for leaf nodes and other types
    def visit_yara_file(self, node) -> None:
        """Visit YARA file node - function call validation is handled at rule level."""

    def visit_import(self, node) -> None:
        """Visit import node - imports are handled by type environment setup."""

    def visit_include(self, node) -> None:
        """Visit include node - includes don't contain function calls to validate."""

    def visit_rule(self, node) -> None:
        """Visit rule node - function call validation is handled in conditions."""

    def visit_tag(self, node) -> None:
        """Visit tag node - tags don't contain function calls."""

    def visit_string_definition(self, node) -> None:
        """Visit string definition node - string definitions don't contain function calls."""

    def visit_plain_string(self, node) -> None:
        """Visit plain string node - plain strings don't contain function calls."""

    def visit_hex_string(self, node) -> None:
        """Visit hex string node - hex strings don't contain function calls."""

    def visit_regex_string(self, node) -> None:
        """Visit regex string node - regex strings don't contain function calls."""

    def visit_string_modifier(self, node) -> None:
        """Visit string modifier node - modifiers don't contain function calls."""

    def visit_hex_token(self, node) -> None:
        """Visit hex token node - hex tokens don't contain function calls."""

    def visit_hex_byte(self, node) -> None:
        """Visit hex byte node - hex bytes don't contain function calls."""

    def visit_hex_wildcard(self, node) -> None:
        """Visit hex wildcard node - wildcards don't contain function calls."""

    def visit_hex_jump(self, node) -> None:
        """Visit hex jump node - jumps don't contain function calls."""

    def visit_hex_alternative(self, node) -> None:
        """Visit hex alternative node - alternatives don't contain function calls."""

    def visit_hex_nibble(self, node) -> None:
        """Visit hex nibble node - nibbles don't contain function calls."""

    def visit_expression(self, node) -> None:
        """Visit expression node - handled by specific expression type visitors."""

    def visit_identifier(self, node) -> None:
        """Visit identifier node - identifiers don't contain function calls."""

    def visit_string_identifier(self, node) -> None:
        """Visit string identifier node - string identifiers don't contain function calls."""

    def visit_string_wildcard(self, node) -> None:
        """Visit string wildcard node - string wildcards don't contain function calls."""

    def visit_string_count(self, node) -> None:
        """Visit string count node - string counts don't contain function calls."""

    def visit_string_offset(self, node) -> None:
        """Visit string offset node - string offsets don't contain function calls."""

    def visit_string_length(self, node) -> None:
        """Visit string length node - string lengths don't contain function calls."""

    def visit_integer_literal(self, node) -> None:
        """Visit integer literal node - integer literals don't contain function calls."""

    def visit_double_literal(self, node) -> None:
        """Visit double literal node - double literals don't contain function calls."""

    def visit_string_literal(self, node) -> None:
        """Visit string literal node - string literals don't contain function calls."""

    def visit_boolean_literal(self, node) -> None:
        """Visit boolean literal node - boolean literals don't contain function calls."""

    def visit_condition(self, node) -> None:
        """Visit condition node - conditions are handled by specific condition visitors."""

    def visit_meta(self, node) -> None:
        """Visit meta node - meta fields don't contain function calls."""

    def visit_module_reference(self, node) -> None:
        """Visit module reference node - module references are handled by member access visitor."""

    def visit_dictionary_access(self, node) -> None:
        """Visit dictionary access node - dictionary access doesn't contain function calls."""

    def visit_comment(self, node) -> None:
        """Visit comment node - comments don't contain function calls."""

    def visit_comment_group(self, node) -> None:
        """Visit comment group node - comment groups don't contain function calls."""

    def visit_defined_expression(self, node) -> None:
        """Visit defined expression node - defined expressions don't contain function calls."""

    def visit_regex_literal(self, node) -> None:
        """Visit regex literal node - regex literals don't contain function calls."""

    def visit_string_operator_expression(self, node) -> None:
        """Visit string operator expression node - string operators don't contain function calls."""

    # Add missing abstract methods for FunctionCallValidator
    def visit_extern_import(self, node) -> None:
        """Visit extern import node - extern imports don't contain function calls."""

    def visit_extern_namespace(self, node) -> None:
        """Visit extern namespace node - extern namespaces don't contain function calls."""

    def visit_extern_rule(self, node) -> None:
        """Visit extern rule node - extern rules don't contain function calls to validate."""

    def visit_extern_rule_reference(self, node) -> None:
        """Visit extern rule reference node - extern rule references don't contain function calls."""

    def visit_in_rule_pragma(self, node) -> None:
        """Visit in-rule pragma node - pragmas don't contain function calls."""

    def visit_pragma(self, node) -> None:
        """Visit pragma node - pragmas don't contain function calls."""

    def visit_pragma_block(self, node) -> None:
        """Visit pragma block node - pragma blocks don't contain function calls."""


class SemanticValidator:
    """Comprehensive semantic validator for YARA AST."""

    def __init__(self) -> None:
        self.module_loader = ModuleLoader()

    def validate(self, ast: YaraFile) -> ValidationResult:
        """Perform complete semantic validation on YARA file."""
        result = ValidationResult()

        # Set up type environment
        env = TypeEnvironment()

        # Process imports first to populate module environment
        for imp in ast.imports:
            alias = imp.alias if imp.alias else imp.module
            env.add_module(alias, imp.module)

        # Add string identifiers to environment for each rule
        for rule in ast.rules:
            for string_def in rule.strings:
                env.add_string(string_def.identifier)

        # 1. Validate string identifier uniqueness per rule
        string_validator = StringIdentifierValidator(result)
        for rule in ast.rules:
            string_validator.visit(rule)

        # 2. Validate function calls and module function existence
        function_validator = FunctionCallValidator(result, env)
        for rule in ast.rules:
            if rule.condition:
                function_validator.visit(rule.condition)

        # 3. Run existing type checker for comprehensive type validation
        type_checker = TypeChecker()
        type_errors = type_checker.check(ast)

        # Convert type checker errors to ValidationError objects
        for error_msg in type_errors:
            result.add_error(
                error_msg,
                suggestion="Check variable types and function signatures",
            )

        return result

    def validate_rule(
        self,
        rule: Rule,
        env: TypeEnvironment | None = None,
    ) -> ValidationResult:
        """Validate a single rule."""
        result = ValidationResult()

        if env is None:
            env = TypeEnvironment()

        # Add rule's string identifiers to environment
        for string_def in rule.strings:
            env.add_string(string_def.identifier)

        # Validate string uniqueness
        string_validator = StringIdentifierValidator(result)
        string_validator.visit(rule)

        # Validate function calls in condition
        if rule.condition:
            function_validator = FunctionCallValidator(result, env)
            function_validator.visit(rule.condition)

        return result

    def validate_expression(
        self,
        expr: Expression,
        env: TypeEnvironment | None = None,
    ) -> ValidationResult:
        """Validate a single expression."""
        result = ValidationResult()

        if env is None:
            env = TypeEnvironment()

        # Validate function calls in expression
        function_validator = FunctionCallValidator(result, env)
        function_validator.visit(expr)

        return result


# Convenience functions for easy usage
def validate_yara_file(ast: YaraFile) -> ValidationResult:
    """Validate YARA file with comprehensive semantic checks."""
    validator = SemanticValidator()
    return validator.validate(ast)


def validate_yara_rule(
    rule: Rule,
    env: TypeEnvironment | None = None,
) -> ValidationResult:
    """Validate a single YARA rule."""
    validator = SemanticValidator()
    return validator.validate_rule(rule, env)


def check_string_uniqueness(rule: Rule) -> list[ValidationError]:
    """Check string identifier uniqueness within a rule."""
    result = ValidationResult()
    validator = StringIdentifierValidator(result)
    validator.visit(rule)
    return result.errors


def check_function_calls(
    expr: Expression,
    env: TypeEnvironment,
) -> list[ValidationError]:
    """Check function calls in an expression."""
    result = ValidationResult()
    validator = FunctionCallValidator(result, env)
    validator.visit(expr)
    return result.errors



"""Document symbols provider for YARA Language Server."""

from __future__ import annotations

from typing import TYPE_CHECKING

from lsprotocol.types import DocumentSymbol, Position, Range, SymbolKind

from yaraast.parser.parser import Parser

if TYPE_CHECKING:
    pass


class SymbolsProvider:
    """Provides document symbols (outline view)."""

    def get_symbols(self, text: str) -> list[DocumentSymbol]:
        """
        Get document symbols for the given YARA file.

        Args:
            text: The YARA source code

        Returns:
            List of document symbols
        """
        symbols = []

        try:
            parser = Parser(text)
            ast = parser.parse()

            lines = text.split("\n")

            # Add imports
            for imp in ast.imports:
                line_num = self._find_line_containing(lines, f'import "{imp.module}"')
                if line_num >= 0:
                    symbols.append(
                        DocumentSymbol(
                            name=f'import "{imp.module}"',
                            kind=SymbolKind.Namespace,
                            range=self._make_range(line_num, 0, line_num, len(lines[line_num])),
                            selection_range=self._make_range(
                                line_num, 0, line_num, len(lines[line_num])
                            ),
                        )
                    )

            # Add rules
            for rule in ast.rules:
                rule_line = self._find_line_containing(lines, f"rule {rule.name}")
                if rule_line < 0:
                    continue

                # Find rule end (closing brace)
                rule_end = self._find_closing_brace(lines, rule_line)

                rule_symbol = DocumentSymbol(
                    name=rule.name,
                    kind=SymbolKind.Class,
                    range=self._make_range(
                        rule_line, 0, rule_end, len(lines[rule_end]) if rule_end < len(lines) else 0
                    ),
                    selection_range=self._make_range(
                        rule_line,
                        lines[rule_line].index(rule.name),
                        rule_line,
                        lines[rule_line].index(rule.name) + len(rule.name),
                    ),
                    children=[],
                )

                # Add meta section
                if rule.meta:
                    meta_line = self._find_line_containing(lines, "meta:", rule_line)
                    if meta_line >= 0:
                        meta_children = []
                        for key, value in rule.meta.items():
                            key_line = self._find_line_containing(lines, f"{key} =", meta_line)
                            if key_line >= 0:
                                meta_children.append(
                                    DocumentSymbol(
                                        name=f"{key} = {value}",
                                        kind=SymbolKind.Property,
                                        range=self._make_range(
                                            key_line, 0, key_line, len(lines[key_line])
                                        ),
                                        selection_range=self._make_range(
                                            key_line, 0, key_line, len(lines[key_line])
                                        ),
                                    )
                                )

                        if meta_children:
                            rule_symbol.children.append(
                                DocumentSymbol(
                                    name="meta",
                                    kind=SymbolKind.Namespace,
                                    range=self._make_range(
                                        meta_line, 0, meta_line, len(lines[meta_line])
                                    ),
                                    selection_range=self._make_range(
                                        meta_line, 0, meta_line, len(lines[meta_line])
                                    ),
                                    children=meta_children,
                                )
                            )

                # Add strings section
                if rule.strings:
                    strings_line = self._find_line_containing(lines, "strings:", rule_line)
                    if strings_line >= 0:
                        string_children = []
                        for string_def in rule.strings:
                            string_line = self._find_line_containing(
                                lines,
                                string_def.identifier,
                                strings_line,
                            )
                            if string_line >= 0:
                                string_children.append(
                                    DocumentSymbol(
                                        name=string_def.identifier,
                                        kind=SymbolKind.String,
                                        range=self._make_range(
                                            string_line, 0, string_line, len(lines[string_line])
                                        ),
                                        selection_range=self._make_range(
                                            string_line,
                                            0,
                                            string_line,
                                            len(lines[string_line]),
                                        ),
                                    )
                                )

                        if string_children:
                            rule_symbol.children.append(
                                DocumentSymbol(
                                    name="strings",
                                    kind=SymbolKind.Namespace,
                                    range=self._make_range(
                                        strings_line, 0, strings_line, len(lines[strings_line])
                                    ),
                                    selection_range=self._make_range(
                                        strings_line,
                                        0,
                                        strings_line,
                                        len(lines[strings_line]),
                                    ),
                                    children=string_children,
                                )
                            )

                # Add condition section
                if rule.condition:
                    condition_line = self._find_line_containing(lines, "condition:", rule_line)
                    if condition_line >= 0:
                        rule_symbol.children.append(
                            DocumentSymbol(
                                name="condition",
                                kind=SymbolKind.Function,
                                range=self._make_range(
                                    condition_line, 0, condition_line, len(lines[condition_line])
                                ),
                                selection_range=self._make_range(
                                    condition_line,
                                    0,
                                    condition_line,
                                    len(lines[condition_line]),
                                ),
                            )
                        )

                symbols.append(rule_symbol)

        except Exception:
            # If parsing fails, return empty symbols
            pass

        return symbols

    def _find_line_containing(
        self,
        lines: list[str],
        text: str,
        start: int = 0,
    ) -> int:
        """Find the line number containing the given text."""
        for i in range(start, len(lines)):
            if text in lines[i]:
                return i
        return -1

    def _find_closing_brace(self, lines: list[str], start: int) -> int:
        """Find the closing brace for a rule."""
        depth = 0
        for i in range(start, len(lines)):
            depth += lines[i].count("{")
            depth -= lines[i].count("}")
            if depth == 0 and "}" in lines[i]:
                return i
        return len(lines) - 1

    def _make_range(self, start_line: int, start_char: int, end_line: int, end_char: int) -> Range:
        """Create an LSP Range."""
        return Range(
            start=Position(line=start_line, character=start_char),
            end=Position(line=end_line, character=end_char),
        )



"""Streaming parser for large YARA files."""

from __future__ import annotations

import mmap
from pathlib import Path
from typing import TYPE_CHECKING, Any

from yaraast.parser import Parser

if TYPE_CHECKING:
    import io
    from collections.abc import Callable, Iterator

    from yaraast.ast.rules import Rule


class StreamingParser:
    """Parse large YARA files efficiently using streaming."""

    def __init__(
        self,
        buffer_size: int = 8192,
        max_memory_mb: int | None = None,
        enable_gc: bool = False,
        progress_callback: Callable | None = None,
    ) -> None:
        """Initialize streaming parser.

        Args:
            buffer_size: Size of read buffer in bytes
            max_memory_mb: Maximum memory usage in MB (ignored for compatibility)
            enable_gc: Enable garbage collection (ignored for compatibility)
            progress_callback: Progress callback function (ignored for compatibility)

        """
        self.buffer_size = buffer_size
        self.max_memory_mb = max_memory_mb
        self.enable_gc = enable_gc
        self.progress_callback = progress_callback
        self.parser = Parser()
        self._cancelled = False
        self._stats = {
            "rules_parsed": 0,
            "bytes_processed": 0,
            "parse_errors": 0,
            "peak_memory_mb": 0,
            "files_processed": 0,
            "files_successful": 0,
            "total_parse_time": 0,
        }

    def parse_file(
        self,
        file_path: str | Path,
        callback: Callable[[Rule], None] | None = None,
    ) -> Iterator[Rule]:
        """Parse a YARA file in streaming fashion.

        Args:
            file_path: Path to YARA file
            callback: Optional callback for each parsed rule

        Yields:
            Parsed rules one at a time

        """
        file_path = Path(file_path)

        with (
            open(file_path, "rb") as f,
            mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ) as mmapped_file,
        ):
            # Use memory mapping for large files
            yield from self._parse_mmap(mmapped_file, callback)

    def parse_stream(
        self,
        stream: io.IOBase,
        callback: Callable[[Rule], None] | None = None,
    ) -> Iterator[Rule]:
        """Parse a stream of YARA content.

        Args:
            stream: Input stream
            callback: Optional callback for each parsed rule

        Yields:
            Parsed rules one at a time

        """
        buffer = ""
        rule_buffer = []
        in_rule = False
        brace_count = 0

        while True:
            chunk = stream.read(self.buffer_size)
            if not chunk:
                break

            if isinstance(chunk, bytes):
                chunk = chunk.decode("utf-8", errors="replace")

            buffer += chunk
            self._stats["bytes_processed"] += len(chunk)

            # Process lines
            lines = buffer.split("\n")
            buffer = lines[-1]  # Keep incomplete line

            for line in lines[:-1]:
                stripped = line.strip()

                # Track rule boundaries
                if stripped.startswith("rule ") and not in_rule:
                    in_rule = True
                    rule_buffer = [line]
                    brace_count = line.count("{") - line.count("}")
                elif in_rule:
                    rule_buffer.append(line)
                    brace_count += line.count("{") - line.count("}")

                    # Complete rule found
                    if brace_count == 0 and "}" in line:
                        rule_text = "\n".join(rule_buffer)
                        rule = self._parse_rule_text(rule_text)
                        if rule:
                            self._stats["rules_parsed"] += 1
                            if callback:
                                callback(rule)
                            yield rule

                        in_rule = False
                        rule_buffer = []

        # Handle any remaining buffer
        if buffer and in_rule:
            rule_buffer.append(buffer)
            rule_text = "\n".join(rule_buffer)
            rule = self._parse_rule_text(rule_text)
            if rule:
                self._stats["rules_parsed"] += 1
                if callback:
                    callback(rule)
                yield rule

    def parse_file_chunked(
        self,
        file_path: str | Path,
        chunk_size: int = 100,
    ) -> Iterator[list[Rule]]:
        """Parse file and yield rules in chunks.

        Args:
            file_path: Path to YARA file
            chunk_size: Number of rules per chunk

        Yields:
            Lists of parsed rules

        """
        chunk = []

        for rule in self.parse_file(file_path):
            chunk.append(rule)
            if len(chunk) >= chunk_size:
                yield chunk
                chunk = []

        # Yield remaining rules
        if chunk:
            yield chunk

    def parse_rules_from_file(self, file_path: Path) -> Iterator[Any]:
        """Parse individual rules from a file (yields one result per rule)."""
        from dataclasses import dataclass
        from enum import Enum

        class ParseStatus(Enum):
            SUCCESS = "success"
            ERROR = "error"

        @dataclass
        class ParseResult:
            file_path: str
            rule_name: str | None
            status: ParseStatus
            error: str | None
            parse_time: float
            rule_count: int
            import_count: int
            ast: Any = None

        try:
            import time

            start_time = time.time()

            for rule in self.parse_file(file_path):
                if self._cancelled:
                    break

                parse_time = time.time() - start_time
                # Create a minimal YaraFile with just this rule
                from yaraast.ast.base import YaraFile

                ast = YaraFile(imports=[], includes=[], rules=[rule])

                yield ParseResult(
                    file_path=str(file_path),
                    rule_name=rule.name if hasattr(rule, "name") else None,
                    status=ParseStatus.SUCCESS,
                    error=None,
                    parse_time=parse_time,
                    rule_count=1,
                    import_count=0,
                    ast=ast,
                )
                start_time = time.time()
        except Exception as e:
            yield ParseResult(
                file_path=str(file_path),
                rule_name=None,
                status=ParseStatus.ERROR,
                error=str(e),
                parse_time=0,
                rule_count=0,
                import_count=0,
                ast=None,
            )

    def parse_files(self, file_paths: list[Path]) -> Iterator[Any]:
        """Parse multiple files (yields one result per file)."""
        from dataclasses import dataclass
        from enum import Enum

        class ParseStatus(Enum):
            SUCCESS = "success"
            ERROR = "error"

        @dataclass
        class ParseResult:
            file_path: str
            rule_name: str | None
            status: ParseStatus
            error: str | None
            parse_time: float
            rule_count: int
            import_count: int
            ast: Any = None

        import time

        for idx, file_path in enumerate(file_paths, 1):
            if self._cancelled:
                break

            try:
                start_time = time.time()
                content = Path(file_path).read_text()
                ast = self.parser.parse(content)
                parse_time = time.time() - start_time

                self._stats["files_processed"] += 1
                self._stats["files_successful"] += 1
                self._stats["total_parse_time"] += parse_time
                self._stats["rules_parsed"] += len(ast.rules)

                # Call progress callback if provided
                if self.progress_callback:
                    self.progress_callback(idx, len(file_paths), str(file_path))

                yield ParseResult(
                    file_path=str(file_path),
                    rule_name=ast.rules[0].name if ast.rules else None,
                    status=ParseStatus.SUCCESS,
                    error=None,
                    parse_time=parse_time,
                    rule_count=len(ast.rules),
                    import_count=len(ast.imports),
                    ast=ast,
                )
            except Exception as e:
                self._stats["files_processed"] += 1
                self._stats["parse_errors"] += 1

                yield ParseResult(
                    file_path=str(file_path),
                    rule_name=None,
                    status=ParseStatus.ERROR,
                    error=str(e),
                    parse_time=0,
                    rule_count=0,
                    import_count=0,
                    ast=None,
                )

    def parse_directory(
        self,
        dir_path: Path,
        pattern: str = "*.yar",
        recursive: bool = False,
    ) -> Iterator[Any]:
        """Parse all files in a directory."""
        files = list(dir_path.rglob(pattern)) if recursive else list(dir_path.glob(pattern))

        yield from self.parse_files(files)

    def get_statistics(self) -> dict[str, Any]:
        """Get parser statistics."""
        return dict(self._stats)

    def cancel(self) -> None:
        """Cancel parsing."""
        self._cancelled = True

    def parse_with_progress(
        self,
        file_path: str | Path,
        progress_callback: Callable[[int, int], None],
    ) -> list[Rule]:
        """Parse file with progress reporting.

        Args:
            file_path: Path to YARA file
            progress_callback: Callback(bytes_processed, total_bytes)

        Returns:
            List of parsed rules

        """
        file_path = Path(file_path)
        file_size = file_path.stat().st_size
        rules = []

        def rule_callback(rule: Rule) -> None:
            rules.append(rule)
            progress_callback(self._stats["bytes_processed"], file_size)

        list(self.parse_file(file_path, rule_callback))
        return rules

    def _parse_mmap(
        self,
        mmapped_file: mmap.mmap,
        callback: Callable[[Rule], None] | None = None,
    ) -> Iterator[Rule]:
        """Parse memory-mapped file content."""
        # Find rule boundaries
        content = mmapped_file.read().decode("utf-8", errors="replace")
        mmapped_file.seek(0)  # Reset position

        # Extract rules by properly tracking brace depth
        import re

        # Find all rule starts
        rule_starts = [(m.start(), m.group(0)) for m in re.finditer(r"rule\s+\w+", content)]

        for i, (start_pos, rule_header) in enumerate(rule_starts):
            # Find the opening brace
            brace_pos = content.find("{", start_pos)
            if brace_pos == -1:
                continue

            # Count braces to find matching closing brace
            brace_count = 1
            pos = brace_pos + 1

            while pos < len(content) and brace_count > 0:
                if content[pos] == "{":
                    brace_count += 1
                elif content[pos] == "}":
                    brace_count -= 1
                pos += 1

            if brace_count == 0:
                # Extract the complete rule text
                rule_text = content[start_pos:pos]
                rule = self._parse_rule_text(rule_text)
                if rule:
                    self._stats["rules_parsed"] += 1
                    if callback:
                        callback(rule)
                    yield rule

    def _parse_rule_text(self, rule_text: str) -> Rule | None:
        """Parse a single rule text."""
        try:
            # Create a minimal YARA file with just this rule
            yara_file = self.parser.parse(rule_text)
            if yara_file.rules:
                return yara_file.rules[0]
        except (ValueError, TypeError, AttributeError):
            self._stats["parse_errors"] += 1

        return None

    def reset_statistics(self) -> None:
        """Reset parsing statistics."""
        self._stats = {
            "rules_parsed": 0,
            "bytes_processed": 0,
            "parse_errors": 0,
        }

    def estimate_memory_usage(self, file_path: str | Path) -> dict[str, Any]:
        """Estimate memory usage for parsing a file.

        Args:
            file_path: Path to YARA file

        Returns:
            Memory usage estimates

        """
        file_path = Path(file_path)
        file_size = file_path.stat().st_size

        # Rough estimates based on experience
        estimated_ast_size = file_size * 3  # AST typically 3x file size
        estimated_peak = file_size * 5  # Peak during parsing

        return {
            "file_size_mb": file_size / 1024 / 1024,
            "estimated_ast_mb": estimated_ast_size / 1024 / 1024,
            "estimated_peak_mb": estimated_peak / 1024 / 1024,
            "streaming_buffer_mb": self.buffer_size / 1024 / 1024,
        }



"""YARA-L semantic validator."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from yaraast.visitor import ASTVisitor

if TYPE_CHECKING:
    from yaraast.yaral.ast_nodes import (
        EventAssignment,
        EventsSection,
        EventStatement,
        EventVariable,
        MatchSection,
        UDMFieldPath,
        YaraLFile,
        YaraLRule,
    )


@dataclass
class ValidationError:
    """Represents a validation error in YARA-L rules."""

    severity: str  # "error", "warning", "info"
    rule_name: str
    section: str
    message: str
    suggestion: str = ""

    def __str__(self) -> str:
        return f"[{self.severity.upper()}] {self.rule_name}/{self.section}: {self.message}"


class YaraLValidator(ASTVisitor[None]):
    """Semantic validator for YARA-L rules."""

    # Valid UDM fields for different event types
    VALID_UDM_FIELDS = {
        "metadata": [
            "event_type",
            "event_timestamp",
            "product_name",
            "vendor_name",
            "description",
            "product_version",
            "product_event_type",
            "ingested_timestamp",
        ],
        "principal": [
            "hostname",
            "ip",
            "port",
            "mac",
            "user",
            "process",
            "file",
            "user_id",
            "email",
            "domain",
            "asset_id",
            "location",
        ],
        "target": [
            "hostname",
            "ip",
            "port",
            "mac",
            "user",
            "process",
            "file",
            "user_id",
            "email",
            "domain",
            "asset_id",
            "url",
            "resource",
        ],
        "observer": ["hostname", "ip", "port", "mac", "asset_id", "location"],
        "src": ["ip", "port", "mac", "hostname", "location", "country", "region"],
        "dst": ["ip", "port", "mac", "hostname", "location", "country", "region"],
        "network": [
            "application_protocol",
            "direction",
            "ip_protocol",
            "received_bytes",
            "sent_bytes",
            "session_duration",
            "session_id",
        ],
        "security_result": [
            "action",
            "category",
            "confidence",
            "severity",
            "rule_name",
            "detection_fields",
            "summary",
            "action_details",
        ],
        "about": ["file", "process", "user", "resource", "labels"],
    }

    # Valid aggregation functions
    VALID_AGGREGATIONS = [
        "count",
        "count_distinct",
        "sum",
        "avg",
        "min",
        "max",
        "array",
        "array_distinct",
        "string_concat",
    ]

    # Valid time units for windows
    VALID_TIME_UNITS = ["s", "m", "h", "d", "seconds", "minutes", "hours", "days"]

    def __init__(self) -> None:
        """Initialize validator."""
        self.errors: list[ValidationError] = []
        self.warnings: list[ValidationError] = []
        self.current_rule: str | None = None
        self.defined_events: set[str] = set()
        self.used_events: set[str] = set()
        self.defined_match_vars: set[str] = set()
        self.used_match_vars: set[str] = set()
        self.defined_outcome_vars: set[str] = set()

    def validate(self, ast: YaraLFile) -> tuple[list[ValidationError], list[ValidationError]]:
        """Validate YARA-L file and return errors and warnings.

        Args:
            ast: YARA-L AST to validate

        Returns:
            Tuple of (errors, warnings)
        """
        self.errors.clear()
        self.warnings.clear()
        self.visit(ast)
        return self.errors, self.warnings

    def _add_error(self, section: str, message: str, suggestion: str = "") -> None:
        """Add validation error."""
        self.errors.append(
            ValidationError(
                severity="error",
                rule_name=self.current_rule or "unknown",
                section=section,
                message=message,
                suggestion=suggestion,
            )
        )

    def _add_warning(self, section: str, message: str, suggestion: str = "") -> None:
        """Add validation warning."""
        self.warnings.append(
            ValidationError(
                severity="warning",
                rule_name=self.current_rule or "unknown",
                section=section,
                message=message,
                suggestion=suggestion,
            )
        )

    def visit_yaral_file(self, node: YaraLFile) -> None:
        """Validate YARA-L file."""
        if not node.rules:
            self._add_warning("file", "Empty YARA-L file")

        rule_names = set()
        for rule in node.rules:
            if rule.name in rule_names:
                self._add_error(
                    "file",
                    f"Duplicate rule name: {rule.name}",
                    "Use unique names for each rule",
                )
            rule_names.add(rule.name)
            self.visit(rule)

    def visit_yaral_rule(self, node: YaraLRule) -> None:
        """Validate YARA-L rule."""
        self.current_rule = node.name
        self.defined_events.clear()
        self.used_events.clear()
        self.defined_match_vars.clear()
        self.used_match_vars.clear()
        self.defined_outcome_vars.clear()

        # Validate rule name
        if not node.name:
            self._add_error("rule", "Rule must have a name")
        elif not node.name[0].isalpha() and node.name[0] != "_":
            self._add_error(
                "rule",
                f"Rule name '{node.name}' must start with letter or underscore",
                "Use valid identifier format",
            )

        # Check required sections
        if not node.events:
            self._add_error(
                "rule",
                "Rule must have an events section",
                "Add 'events:' section to define event patterns",
            )

        if not node.condition:
            self._add_error(
                "rule",
                "Rule must have a condition section",
                "Add 'condition:' section to define matching conditions",
            )

        # Validate sections if present
        if node.meta:
            self._validate_meta_section(node.meta)

        if node.events:
            self.visit(node.events)

        if node.match:
            self._validate_match_section(node.match)

        if node.condition:
            self._validate_condition_section(node.condition)

        if node.outcome:
            self._validate_outcome_section(node.outcome)

        if node.options:
            self._validate_options_section(node.options)

        # Cross-section validation
        self._validate_cross_sections()

    def visit_events_section(self, node: EventsSection) -> None:
        """Validate events section."""
        if not node.statements:
            self._add_error(
                "events",
                "Events section cannot be empty",
                "Add at least one event statement",
            )

        for statement in node.statements:
            self.visit(statement)

    def visit_event_statement(self, node: EventStatement) -> None:
        """Validate event statement."""
        if node.event:
            self.visit(node.event)

        if not node.assignments:
            self._add_warning(
                "events",
                f"Event {node.event.name if node.event else 'unknown'} has no field assignments",
                "Add field assignments to constrain the event",
            )

        for assignment in node.assignments:
            self.visit(assignment)

    def visit_event_variable(self, node: EventVariable) -> None:
        """Validate event variable."""
        if not node.name.startswith("$"):
            self._add_error(
                "events",
                f"Event variable '{node.name}' must start with $",
                "Use format like $e, $e1, $event",
            )

        if node.name in self.defined_events:
            self._add_error(
                "events",
                f"Duplicate event variable: {node.name}",
                "Use unique event variable names",
            )

        self.defined_events.add(node.name)

    def visit_event_assignment(self, node: EventAssignment) -> None:
        """Validate event assignment."""
        # Validate field path
        if node.field_path:
            self._validate_udm_field_path(node.field_path)

        # Validate operator
        valid_operators = ["=", "!=", ">", "<", ">=", "<=", "=~", "!~"]
        if node.operator not in valid_operators:
            self._add_error(
                "events",
                f"Invalid operator '{node.operator}' in event assignment",
                f"Use one of: {', '.join(valid_operators)}",
            )

        # Validate regex operators
        if node.operator in ["=~", "!~"] and (
            not isinstance(node.value, str)
            or not (node.value.startswith("/") and node.value.endswith("/"))
        ):
            self._add_warning(
                "events",
                f"Regex operator {node.operator} should be used with regex pattern",
                "Use format: /pattern/",
            )

    def _validate_udm_field_path(self, node: UDMFieldPath) -> None:
        """Validate UDM field path."""
        if not node.parts:
            self._add_error("events", "Empty UDM field path")
            return

        # Check if first part is valid UDM namespace
        namespace = node.parts[0]
        if namespace not in self.VALID_UDM_FIELDS:
            self._add_warning(
                "events",
                f"Unknown UDM namespace: {namespace}",
                f"Valid namespaces: {', '.join(self.VALID_UDM_FIELDS.keys())}",
            )
        elif len(node.parts) > 1:
            # Check if field is valid for namespace
            field = node.parts[1]
            valid_fields = self.VALID_UDM_FIELDS[namespace]
            if field not in valid_fields:
                self._add_warning(
                    "events",
                    f"Unknown field '{field}' for namespace '{namespace}'",
                    f"Valid fields: {', '.join(valid_fields[:5])}...",
                )

    def _validate_meta_section(self, node: Any) -> None:
        """Validate meta section."""
        required_meta = ["author", "description"]
        found_keys = set()

        if hasattr(node, "entries"):
            for entry in node.entries:
                found_keys.add(entry.key)

                # Validate specific meta fields
                if entry.key == "severity" and isinstance(entry.value, str):
                    valid_severities = [
                        "informational",
                        "low",
                        "medium",
                        "high",
                        "critical",
                    ]
                    if entry.value.lower() not in valid_severities:
                        self._add_warning(
                            "meta",
                            f"Invalid severity value: {entry.value}",
                            f"Use one of: {', '.join(valid_severities)}",
                        )

        # Check for recommended meta fields
        for key in required_meta:
            if key not in found_keys:
                self._add_warning(
                    "meta",
                    f"Missing recommended meta field: {key}",
                    f"Add '{key}' to meta section",
                )

    def _validate_match_section(self, node: MatchSection) -> None:
        """Validate match section."""
        if not node.variables:
            self._add_warning(
                "match",
                "Match section has no variables",
                "Define match variables for correlation",
            )

        for var in node.variables:
            if var.name in self.defined_match_vars:
                self._add_error(
                    "match",
                    f"Duplicate match variable: {var.name}",
                    "Use unique match variable names",
                )
            self.defined_match_vars.add(var.name)

        # Validate time window
        if (
            node.time_window
            and hasattr(node.time_window, "unit")
            and node.time_window.unit not in self.VALID_TIME_UNITS
        ):
            self._add_error(
                "match",
                f"Invalid time unit: {node.time_window.unit}",
                f"Use one of: {', '.join(self.VALID_TIME_UNITS)}",
            )

            if hasattr(node.time_window, "duration"):
                if node.time_window.duration <= 0:
                    self._add_error(
                        "match",
                        "Time window duration must be positive",
                        "Use positive duration value",
                    )
                elif node.time_window.duration > 30 and node.time_window.unit in [
                    "d",
                    "days",
                ]:
                    self._add_warning(
                        "match",
                        f"Large time window: {node.time_window.duration} {node.time_window.unit}",
                        "Consider using smaller time windows for better performance",
                    )

    def _validate_condition_section(self, node: Any) -> None:
        """Validate condition section."""
        if not hasattr(node, "expression") or not node.expression:
            self._add_error(
                "condition",
                "Condition section cannot be empty",
                "Add condition expression",
            )

        # Check for event references
        # This would need deeper AST traversal in a real implementation

    def _validate_outcome_section(self, node: Any) -> None:
        """Validate outcome section."""
        if hasattr(node, "assignments"):
            for assignment in node.assignments:
                var_name = assignment.variable
                if var_name in self.defined_outcome_vars:
                    self._add_error(
                        "outcome",
                        f"Duplicate outcome variable: {var_name}",
                        "Use unique outcome variable names",
                    )
                self.defined_outcome_vars.add(var_name)

                # Check for reserved names
                reserved = ["risk_score", "severity", "confidence"]
                if var_name not in reserved and not var_name.startswith("$"):
                    self._add_warning(
                        "outcome",
                        f"Outcome variable '{var_name}' should start with $ or use reserved name",
                        f"Reserved names: {', '.join(reserved)}",
                    )
        elif hasattr(node, "variables"):
            for var_name in node.variables:
                if var_name in self.defined_outcome_vars:
                    self._add_error(
                        "outcome",
                        f"Duplicate outcome variable: {var_name}",
                        "Use unique outcome variable names",
                    )
                self.defined_outcome_vars.add(var_name)

                # Check for reserved names
                reserved = ["risk_score", "severity", "confidence"]
                if var_name not in reserved and not var_name.startswith("$"):
                    self._add_warning(
                        "outcome",
                        f"Outcome variable '{var_name}' should start with $ or use reserved name",
                        f"Reserved names: {', '.join(reserved)}",
                    )

    def _validate_options_section(self, node: Any) -> None:
        """Validate options section."""
        valid_options = [
            "allow_zero_values",
            "case_sensitive",
            "max_events",
            "max_matches",
            "timeout",
            "output_format",
        ]

        if hasattr(node, "options"):
            for key in node.options:
                if key not in valid_options:
                    self._add_warning(
                        "options",
                        f"Unknown option: {key}",
                        f"Valid options: {', '.join(valid_options)}",
                    )

    def _validate_cross_sections(self) -> None:
        """Validate cross-section references."""
        # Check if all used events are defined
        # Note: We filter out variables that might be outcome variables
        undefined_events = self.used_events - self.defined_events
        # Remove variables that are defined in outcome section
        undefined_events = {
            e
            for e in undefined_events
            if e not in self.defined_outcome_vars and f"${e}" not in self.defined_outcome_vars
        }
        for event in undefined_events:
            self._add_error(
                "condition",
                f"Undefined event variable: {event}",
                "Define event in events section",
            )

        # Check if all defined events are used
        unused_events = self.defined_events - self.used_events
        for event in unused_events:
            self._add_warning(
                "events",
                f"Unused event variable: {event}",
                "Remove unused event or use it in condition",
            )

        # Check match variables usage
        unused_match_vars = self.defined_match_vars - self.used_match_vars
        for var in unused_match_vars:
            self._add_warning(
                "match",
                f"Unused match variable: {var}",
                "Remove unused variable or use it in outcome",
            )



"""Hover provider for YARA Language Server."""

from __future__ import annotations

from typing import TYPE_CHECKING

from lsprotocol.types import Hover, MarkupContent, MarkupKind, Position

from yaraast.lsp.utils import get_word_at_position
from yaraast.parser.parser import Parser
from yaraast.types.module_loader import ModuleLoader

if TYPE_CHECKING:
    pass


class HoverProvider:
    """Provides hover information for YARA symbols."""

    def __init__(self) -> None:
        self.module_loader = ModuleLoader()
        self._init_documentation()

    def _init_documentation(self) -> None:
        """Initialize documentation for YARA keywords and built-ins."""
        self.keyword_docs = {
            "rule": "Defines a YARA rule. Each rule has a name, optional tags, meta section, strings section, and condition section.",
            "private": "Makes a rule private. Private rules are not reported by YARA, but can be used by other rules.",
            "global": "Makes a rule global. Global rules are applied to all files.",
            "meta": "Metadata section containing key-value pairs describing the rule.",
            "strings": "Section where string patterns are defined using $identifier syntax.",
            "condition": "Boolean expression that determines if the rule matches.",
            "import": "Import a YARA module to access additional functions and data structures.",
            "include": "Include another YARA file.",
            "and": "Logical AND operator",
            "or": "Logical OR operator",
            "not": "Logical NOT operator",
            "all": "Quantifier meaning 'all of'",
            "any": "Quantifier meaning 'any of'",
            "of": "Used in quantifier expressions like 'any of them'",
            "them": "Refers to all string identifiers in the current rule",
            "for": "Loop construct for iterating over collections",
            "in": "Membership operator or part of for loop",
            "at": "Tests if a string appears at a specific offset",
            "filesize": "Built-in variable containing the size of the file being scanned in bytes",
            "entrypoint": "Built-in variable containing the entry point address of PE files",
            "true": "Boolean true value",
            "false": "Boolean false value",
            "defined": "Tests if an expression is defined (useful for optional module fields)",
        }

        self.builtin_docs = {
            "uint8": "```yara\nuint8(offset) -> integer\n```\n\nReads an unsigned 8-bit integer at the given offset.",
            "uint16": "```yara\nuint16(offset) -> integer\n```\n\nReads an unsigned 16-bit integer at the given offset (little-endian).",
            "uint32": "```yara\nuint32(offset) -> integer\n```\n\nReads an unsigned 32-bit integer at the given offset (little-endian).",
            "uint16le": "```yara\nuint16le(offset) -> integer\n```\n\nReads an unsigned 16-bit integer at the given offset (little-endian).",
            "uint32le": "```yara\nuint32le(offset) -> integer\n```\n\nReads an unsigned 32-bit integer at the given offset (little-endian).",
            "uint16be": "```yara\nuint16be(offset) -> integer\n```\n\nReads an unsigned 16-bit integer at the given offset (big-endian).",
            "uint32be": "```yara\nuint32be(offset) -> integer\n```\n\nReads an unsigned 32-bit integer at the given offset (big-endian).",
            "int8": "```yara\nint8(offset) -> integer\n```\n\nReads a signed 8-bit integer at the given offset.",
            "int16": "```yara\nint16(offset) -> integer\n```\n\nReads a signed 16-bit integer at the given offset (little-endian).",
            "int32": "```yara\nint32(offset) -> integer\n```\n\nReads a signed 32-bit integer at the given offset (little-endian).",
        }

        self.module_docs = {
            "pe": "PE file format module. Provides access to PE headers, sections, imports, exports, and resources.",
            "elf": "ELF file format module. Provides access to ELF headers, sections, and segments.",
            "math": "Mathematical operations module. Provides entropy calculation and other math functions.",
            "hash": "Hash calculation module. Provides MD5, SHA1, SHA256, and checksum functions.",
            "dotnet": ".NET module. Provides access to .NET assembly metadata.",
            "time": "Time module. Provides functions for time-based operations.",
            "magic": "Magic module. Provides file type identification.",
            "console": "Console module. Provides console output for debugging.",
            "cuckoo": "Cuckoo sandbox integration module.",
            "string": "String manipulation module.",
        }

    def get_hover(self, text: str, position: Position) -> Hover | None:
        """
        Get hover information for the symbol at the given position.

        Args:
            text: The YARA source code
            position: The cursor position

        Returns:
            Hover information or None
        """
        word, word_range = get_word_at_position(text, position)

        if not word:
            return None

        # Check keywords
        if word in self.keyword_docs:
            return Hover(
                contents=MarkupContent(
                    kind=MarkupKind.Markdown,
                    value=f"**{word}** (keyword)\n\n{self.keyword_docs[word]}",
                ),
                range=word_range,
            )

        # Check built-in functions
        if word in self.builtin_docs:
            return Hover(
                contents=MarkupContent(
                    kind=MarkupKind.Markdown,
                    value=f"**{word}** (built-in function)\n\n{self.builtin_docs[word]}",
                ),
                range=word_range,
            )

        # Check modules
        if word in self.module_docs:
            return Hover(
                contents=MarkupContent(
                    kind=MarkupKind.Markdown,
                    value=f"**{word}** (module)\n\n{self.module_docs[word]}",
                ),
                range=word_range,
            )

        # Check module members (e.g., pe.imphash)
        if "." in word:
            parts = word.split(".")
            if len(parts) == 2:
                module_name, member_name = parts
                return self._get_module_member_hover(module_name, member_name, word_range)

        # Check string identifiers
        if word.startswith("$"):
            return self._get_string_identifier_hover(text, word, word_range)

        # Check rule names
        rule_hover = self._get_rule_hover(text, word, word_range)
        if rule_hover:
            return rule_hover

        return None

    def _get_module_member_hover(
        self,
        module_name: str,
        member_name: str,
        word_range,
    ) -> Hover | None:
        """Get hover for a module member (field or function)."""
        module_def = self.module_loader.get_module(module_name)
        if not module_def:
            return None

        # Check functions
        if member_name in module_def.functions:
            func_def = module_def.functions[member_name]
            params = ", ".join(f"{p[0]}: {p[1]}" for p in func_def.parameters)
            signature = f"{module_name}.{member_name}({params}) -> {func_def.return_type}"
            doc = f"**{member_name}** (function)\n\n```yara\n{signature}\n```"

            if func_def.description:
                doc += f"\n\n{func_def.description}"

            return Hover(
                contents=MarkupContent(kind=MarkupKind.Markdown, value=doc),
                range=word_range,
            )

        # Check fields
        if member_name in module_def.fields:
            field_def = module_def.fields[member_name]
            doc = f"**{member_name}** (field)\n\n```yara\n{module_name}.{member_name}: {field_def.type}\n```"

            if field_def.description:
                doc += f"\n\n{field_def.description}"

            return Hover(
                contents=MarkupContent(kind=MarkupKind.Markdown, value=doc),
                range=word_range,
            )

        return None

    def _get_string_identifier_hover(
        self,
        text: str,
        identifier: str,
        word_range,
    ) -> Hover | None:
        """Get hover for a string identifier."""
        try:
            parser = Parser(text)
            ast = parser.parse()

            # Find the string definition
            for rule in ast.rules:
                for string_def in rule.strings:
                    if string_def.identifier == identifier:
                        # Get string value
                        if hasattr(string_def, "value"):
                            value = string_def.value
                            string_type = "text string"
                        elif hasattr(string_def, "regex"):
                            value = string_def.regex
                            string_type = "regex"
                        elif hasattr(string_def, "tokens"):
                            value = "<hex pattern>"
                            string_type = "hex string"
                        else:
                            value = "<unknown>"
                            string_type = "string"

                        # Get modifiers
                        modifiers = []
                        if hasattr(string_def, "modifiers"):
                            modifiers = [m.name for m in string_def.modifiers]

                        doc = f"**{identifier}** ({string_type})\n\n"
                        doc += f"```\n{value}\n```"

                        if modifiers:
                            doc += f"\n\nModifiers: {', '.join(modifiers)}"

                        return Hover(
                            contents=MarkupContent(kind=MarkupKind.Markdown, value=doc),
                            range=word_range,
                        )

        except Exception:
            pass

        # Generic string identifier hover
        return Hover(
            contents=MarkupContent(
                kind=MarkupKind.Markdown,
                value=f"**{identifier}** (string identifier)\n\nString pattern defined in the strings section.",
            ),
            range=word_range,
        )

    def _get_rule_hover(self, text: str, rule_name: str, word_range) -> Hover | None:
        """Get hover for a rule name."""
        try:
            parser = Parser(text)
            ast = parser.parse()

            for rule in ast.rules:
                if rule.name == rule_name:
                    # Build rule documentation
                    doc = f"**{rule_name}** (rule)"

                    if rule.modifiers:
                        doc += f" [{', '.join(rule.modifiers)}]"

                    if rule.tags:
                        tags = [tag.name for tag in rule.tags]
                        doc += f"\n\nTags: {', '.join(tags)}"

                    if rule.meta:
                        doc += "\n\n**Metadata:**\n"
                        for key, value in rule.meta.items():
                            doc += f"- {key}: {value}\n"

                    if rule.strings:
                        doc += f"\n\n**Strings:** {len(rule.strings)} defined"

                    return Hover(
                        contents=MarkupContent(kind=MarkupKind.Markdown, value=doc),
                        range=word_range,
                    )

        except Exception:
            pass

        return None



"""Protobuf serialization for YARA AST."""

from __future__ import annotations

import time
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any

from yaraast.visitor import ASTVisitor

if TYPE_CHECKING:
    from yaraast.ast.base import YaraFile

# Suppress protobuf version warning before import
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    message=".*Protobuf gencode version.*",
    module="google.protobuf.runtime_version",
)

try:
    from yaraast.serialization import yara_ast_pb2
except ImportError:
    # Fallback if protobuf compilation failed
    yara_ast_pb2 = None


class ProtobufSerializer(ASTVisitor[Any]):
    """Protobuf serializer for YARA AST with efficient binary format."""

    def __init__(self, include_metadata: bool = True) -> None:
        if yara_ast_pb2 is None:
            msg = "Protobuf schema not compiled. Run: protoc --python_out=. yara_ast.proto"
            raise ImportError(
                msg,
            )

        self.include_metadata = include_metadata

    def serialize(self, ast: YaraFile, output_path: str | Path | None = None) -> bytes:
        """Serialize AST to Protobuf binary format."""
        pb_yara_file = self._ast_to_protobuf(ast)
        binary_data = pb_yara_file.SerializeToString()

        if output_path:
            with Path(output_path).open("wb") as f:
                f.write(binary_data)

        return binary_data

    def serialize_text(
        self,
        ast: YaraFile,
        output_path: str | Path | None = None,
    ) -> str:
        """Serialize AST to Protobuf text format (for debugging)."""
        pb_yara_file = self._ast_to_protobuf(ast)
        text_data = str(pb_yara_file)

        if output_path:
            with Path(output_path).open("w", encoding="utf-8") as f:
                f.write(text_data)

        return text_data

    def deserialize(
        self,
        binary_data: bytes | None = None,
        input_path: str | Path | None = None,
    ) -> YaraFile:
        """Deserialize Protobuf binary to AST."""
        if input_path:
            with Path(input_path).open("rb") as f:
                binary_data = f.read()

        if not binary_data:
            msg = "No binary data provided"
            raise ValueError(msg)

        pb_yara_file = yara_ast_pb2.YaraFile()
        pb_yara_file.ParseFromString(binary_data)

        return self._protobuf_to_ast(pb_yara_file)

    def _ast_to_protobuf(self, ast: YaraFile) -> yara_ast_pb2.YaraFile:
        """Convert AST to Protobuf message."""
        pb_file = yara_ast_pb2.YaraFile()

        # Convert imports
        for imp in ast.imports:
            pb_import = pb_file.imports.add()
            pb_import.module = imp.module
            if hasattr(imp, "alias") and imp.alias:
                pb_import.alias = imp.alias

        # Convert includes
        for inc in ast.includes:
            pb_include = pb_file.includes.add()
            pb_include.path = inc.path

        # Convert rules
        for rule in ast.rules:
            pb_rule = pb_file.rules.add()
            self._convert_rule_to_protobuf(rule, pb_rule)

        # Add metadata
        if self.include_metadata:
            pb_file.metadata.format = "yaraast-protobuf"
            pb_file.metadata.version = "1.0"
            pb_file.metadata.ast_type = "YaraFile"
            pb_file.metadata.rules_count = len(ast.rules)
            pb_file.metadata.imports_count = len(ast.imports)
            pb_file.metadata.includes_count = len(ast.includes)
            pb_file.metadata.timestamp = int(time.time())

        return pb_file

    def _convert_rule_to_protobuf(self, rule, pb_rule) -> None:
        """Convert rule AST to Protobuf."""
        pb_rule.name = rule.name
        pb_rule.modifiers.extend(rule.modifiers)

        # Convert tags
        for tag in rule.tags:
            pb_tag = pb_rule.tags.add()
            pb_tag.name = tag.name

        # Convert meta
        for key, value in rule.meta.items():
            meta_val = pb_rule.meta[key]
            if isinstance(value, str):
                meta_val.string_value = value
            elif isinstance(value, int):
                meta_val.int_value = value
            elif isinstance(value, bool):
                meta_val.bool_value = value
            elif isinstance(value, float):
                meta_val.double_value = value

        # Convert strings
        for string_def in rule.strings:
            pb_string = pb_rule.strings.add()
            pb_string.identifier = string_def.identifier
            self._convert_string_to_protobuf(string_def, pb_string)

        # Convert condition
        if rule.condition:
            self._convert_expression_to_protobuf(rule.condition, pb_rule.condition)

    def _convert_string_to_protobuf(self, string_def, pb_string) -> None:
        """Convert string definition to Protobuf."""
        from yaraast.ast.strings import HexString, PlainString, RegexString

        if isinstance(string_def, PlainString):
            pb_string.plain.value = string_def.value
            for mod in string_def.modifiers:
                pb_mod = pb_string.plain.modifiers.add()
                pb_mod.name = mod.name
                if mod.value:
                    pb_mod.value = mod.value

        elif isinstance(string_def, HexString):
            for token in string_def.tokens:
                pb_token = pb_string.hex.tokens.add()
                self._convert_hex_token_to_protobuf(token, pb_token)

            for mod in string_def.modifiers:
                pb_mod = pb_string.hex.modifiers.add()
                pb_mod.name = mod.name
                if mod.value:
                    pb_mod.value = mod.value

        elif isinstance(string_def, RegexString):
            pb_string.regex.regex = string_def.regex
            for mod in string_def.modifiers:
                pb_mod = pb_string.regex.modifiers.add()
                pb_mod.name = mod.name
                if mod.value:
                    pb_mod.value = mod.value

    def _convert_hex_token_to_protobuf(self, token, pb_token) -> None:
        """Convert hex token to Protobuf."""
        from yaraast.ast.strings import HexByte, HexJump, HexWildcard
        from yaraast.builder.hex_string_builder import HexNibble

        if isinstance(token, HexByte):
            pb_token.byte.value = str(token.value)
        elif isinstance(token, HexWildcard):
            pb_token.wildcard.CopyFrom(yara_ast_pb2.HexWildcard())
        elif isinstance(token, HexJump):
            pb_token.jump.min_jump = token.min_jump or 0
            pb_token.jump.max_jump = token.max_jump or 0
        elif isinstance(token, HexNibble):
            pb_token.nibble.high = token.high
            pb_token.nibble.value = token.value
        # Note: HexAlternative would need more complex handling

    def _convert_expression_to_protobuf(self, expr, pb_expr) -> None:
        """Convert expression to Protobuf."""
        from yaraast.ast.expressions import (
            BinaryExpression,
            BooleanLiteral,
            DoubleLiteral,
            Identifier,
            IntegerLiteral,
            StringCount,
            StringIdentifier,
            StringLiteral,
            UnaryExpression,
        )

        if isinstance(expr, Identifier):
            pb_expr.identifier.name = expr.name
        elif isinstance(expr, StringIdentifier):
            pb_expr.string_identifier.name = expr.name
        elif isinstance(expr, StringCount):
            pb_expr.string_count.string_id = expr.string_id
        elif isinstance(expr, IntegerLiteral):
            pb_expr.integer_literal.value = expr.value
        elif isinstance(expr, DoubleLiteral):
            pb_expr.double_literal.value = expr.value
        elif isinstance(expr, StringLiteral):
            pb_expr.string_literal.value = expr.value
        elif isinstance(expr, BooleanLiteral):
            pb_expr.boolean_literal.value = expr.value
        elif isinstance(expr, BinaryExpression):
            pb_expr.binary_expression.operator = expr.operator
            self._convert_expression_to_protobuf(
                expr.left,
                pb_expr.binary_expression.left,
            )
            self._convert_expression_to_protobuf(
                expr.right,
                pb_expr.binary_expression.right,
            )
        elif isinstance(expr, UnaryExpression):
            pb_expr.unary_expression.operator = expr.operator
            self._convert_expression_to_protobuf(
                expr.operand,
                pb_expr.unary_expression.operand,
            )
        # Add more expression types as needed

    def _protobuf_to_ast(self, pb_file: yara_ast_pb2.YaraFile) -> YaraFile:
        """Convert Protobuf message to AST."""
        from yaraast.ast.base import YaraFile
        from yaraast.ast.expressions import BooleanLiteral
        from yaraast.ast.rules import Import, Include, Rule

        # Reconstruct imports
        imports = []
        for pb_import in pb_file.imports:
            imports.append(
                Import(
                    module=pb_import.module,
                    alias=pb_import.alias if pb_import.alias else None,
                ),
            )

        # Reconstruct includes
        includes = []
        for pb_include in pb_file.includes:
            includes.append(Include(path=pb_include.path))

        # Reconstruct rules (basic)
        rules = []
        for pb_rule in pb_file.rules:
            # For now, create basic rules with boolean conditions
            rule = Rule(
                name=pb_rule.name,
                modifiers=list(pb_rule.modifiers),
                tags=[],
                meta={},
                strings=[],
                condition=BooleanLiteral(value=True),  # Placeholder condition
            )
            rules.append(rule)

        return YaraFile(imports=imports, includes=includes, rules=rules)

    def get_serialization_stats(self, ast: YaraFile) -> dict[str, Any]:
        """Get statistics about the serialization."""
        pb_file = self._ast_to_protobuf(ast)
        binary_size = len(pb_file.SerializeToString())
        text_size = len(str(pb_file))

        return {
            "binary_size_bytes": binary_size,
            "text_size_bytes": text_size,
            "compression_ratio": text_size / binary_size if binary_size > 0 else 0,
            "rules_count": len(ast.rules),
            "imports_count": len(ast.imports),
            "includes_count": len(ast.includes),
        }

    # Required visitor methods (simplified for now)
    def visit_yara_file(self, node) -> Any:
        return None

    def visit_import(self, node) -> Any:
        return None

    def visit_include(self, node) -> Any:
        return None

    def visit_rule(self, node) -> Any:
        return None

    def visit_tag(self, node) -> Any:
        return None

    def visit_string_definition(self, node) -> Any:
        return None

    def visit_plain_string(self, node) -> Any:
        return None

    def visit_hex_string(self, node) -> Any:
        return None

    def visit_regex_string(self, node) -> Any:
        return None

    def visit_string_modifier(self, node) -> Any:
        return None

    def visit_hex_token(self, node) -> Any:
        return None

    def visit_hex_byte(self, node) -> Any:
        return None

    def visit_hex_wildcard(self, node) -> Any:
        return None

    def visit_hex_jump(self, node) -> Any:
        return None

    def visit_hex_alternative(self, node) -> Any:
        return None

    def visit_hex_nibble(self, node) -> Any:
        return None

    def visit_expression(self, node) -> Any:
        return None

    def visit_identifier(self, node) -> Any:
        return None

    def visit_string_identifier(self, node) -> Any:
        return None

    def visit_string_count(self, node) -> Any:
        return None

    def visit_string_offset(self, node) -> Any:
        return None

    def visit_string_length(self, node) -> Any:
        return None

    def visit_integer_literal(self, node) -> Any:
        return None

    def visit_double_literal(self, node) -> Any:
        return None

    def visit_string_literal(self, node) -> Any:
        return None

    def visit_regex_literal(self, node) -> Any:
        return None

    def visit_boolean_literal(self, node) -> Any:
        return None

    def visit_binary_expression(self, node) -> Any:
        return None

    def visit_unary_expression(self, node) -> Any:
        return None

    def visit_parentheses_expression(self, node) -> Any:
        return None

    def visit_set_expression(self, node) -> Any:
        return None

    def visit_range_expression(self, node) -> Any:
        return None

    def visit_function_call(self, node) -> Any:
        return None

    def visit_array_access(self, node) -> Any:
        return None

    def visit_member_access(self, node) -> Any:
        return None

    def visit_condition(self, node) -> Any:
        return None

    def visit_for_expression(self, node) -> Any:
        return None

    def visit_for_of_expression(self, node) -> Any:
        return None

    def visit_at_expression(self, node) -> Any:
        return None

    def visit_in_expression(self, node) -> Any:
        return None

    def visit_of_expression(self, node) -> Any:
        return None

    def visit_meta(self, node) -> Any:
        return None

    def visit_module_reference(self, node) -> Any:
        return None

    def visit_dictionary_access(self, node) -> Any:
        return None

    def visit_comment(self, node) -> Any:
        return None

    def visit_comment_group(self, node) -> Any:
        return None

    def visit_defined_expression(self, node) -> Any:
        return None

    def visit_string_operator_expression(self, node) -> Any:
        return None

    def visit_extern_import(self, node) -> Any:
        """Visit ExternImport node."""
        return None

    def visit_string_wildcard(self, node) -> Any:
        """Visit StringWildcard node."""
        return None

    def visit_extern_namespace(self, node) -> Any:
        """Visit ExternNamespace node."""
        return None

    def visit_extern_rule(self, node) -> Any:
        """Visit ExternRule node."""
        return None

    def visit_extern_rule_reference(self, node) -> Any:
        """Visit ExternRuleReference node."""
        return None

    def visit_in_rule_pragma(self, node) -> Any:
        """Visit InRulePragma node."""
        return None

    def visit_pragma(self, node) -> Any:
        """Visit Pragma node."""
        return None

    def visit_pragma_block(self, node) -> Any:
        """Visit PragmaBlock node."""
        return None



"""Dependency analyzer for YARA rules."""

# type: ignore  # Analysis code allows gradual typing

from collections import defaultdict

from yaraast.ast.base import YaraFile
from yaraast.ast.conditions import (
    AtExpression,
    ForExpression,
    ForOfExpression,
    InExpression,
    OfExpression,
)
from yaraast.ast.expressions import (
    ArrayAccess,
    BinaryExpression,
    BooleanLiteral,
    DoubleLiteral,
    Expression,
    FunctionCall,
    Identifier,
    IntegerLiteral,
    MemberAccess,
    ParenthesesExpression,
    RangeExpression,
    SetExpression,
    StringCount,
    StringIdentifier,
    StringLength,
    StringLiteral,
    StringOffset,
    UnaryExpression,
)
from yaraast.ast.meta import Meta
from yaraast.ast.rules import Import, Include, Rule, Tag
from yaraast.ast.strings import (
    HexAlternative,
    HexByte,
    HexJump,
    HexNibble,
    HexString,
    HexToken,
    HexWildcard,
    PlainString,
    RegexString,
    StringDefinition,
    StringModifier,
)
from yaraast.visitor import ASTVisitor


class DependencyAnalyzer(ASTVisitor[None]):
    """Analyze dependencies between YARA rules."""

    def __init__(self) -> None:
        self.rule_names: set[str] = set()
        self.dependencies: dict[str, set[str]] = defaultdict(
            set,
        )  # rule -> rules it depends on
        self.current_rule: str | None = None
        self.imported_modules: set[str] = set()
        self.included_files: set[str] = set()

    def analyze(self, yara_file: YaraFile) -> dict[str, any]:
        """Analyze dependencies in YARA file."""
        self.rule_names.clear()
        self.dependencies.clear()
        self.imported_modules.clear()
        self.included_files.clear()

        # First pass: collect all rule names
        for rule in yara_file.rules:
            self.rule_names.add(rule.name)

        # Visit imports and includes
        for imp in yara_file.imports:
            self.visit(imp)

        for inc in yara_file.includes:
            self.visit(inc)

        # Second pass: analyze dependencies
        for rule in yara_file.rules:
            self.visit(rule)

        return {
            "rules": list(self.rule_names),
            "dependencies": {rule: list(deps) for rule, deps in self.dependencies.items()},
            "dependency_graph": self._build_dependency_graph(),
            "circular_dependencies": self._find_circular_dependencies(),
            "dependency_order": self._topological_sort(),
            "imported_modules": list(self.imported_modules),
            "included_files": list(self.included_files),
        }

    def get_dependencies(self, rule_name: str) -> list[str]:
        """Get direct dependencies of a rule."""
        return list(self.dependencies.get(rule_name, set()))

    def get_dependents(self, rule_name: str) -> list[str]:
        """Get rules that depend on the given rule."""
        dependents = []
        for rule, deps in self.dependencies.items():
            if rule_name in deps:
                dependents.append(rule)
        return dependents

    def get_transitive_dependencies(self, rule_name: str) -> set[str]:
        """Get all transitive dependencies of a rule."""
        visited = set()
        to_visit = [rule_name]

        while to_visit:
            current = to_visit.pop()
            if current in visited:
                continue

            visited.add(current)
            deps = self.dependencies.get(current, set())
            to_visit.extend(deps - visited)

        visited.remove(rule_name)  # Don't include self
        return visited

    def _build_dependency_graph(self) -> dict[str, dict[str, list[str]]]:
        """Build a dependency graph."""
        graph = {}

        for rule in self.rule_names:
            deps = self.dependencies.get(rule, set())
            dependents = self.get_dependents(rule)
            transitive = self.get_transitive_dependencies(rule)

            graph[rule] = {
                "depends_on": list(deps),
                "depended_by": dependents,
                "transitive_dependencies": list(transitive),
                "is_independent": len(deps) == 0 and len(dependents) == 0,
            }

        return graph

    def _find_circular_dependencies(self) -> list[list[str]]:
        """Find circular dependencies using DFS."""
        dfs_state = self._init_dfs_state()
        cycles = []

        for rule in self.rule_names:
            if dfs_state["color"][rule] == dfs_state["white"]:
                self._dfs_cycle_detection(rule, dfs_state, cycles)

        return self._remove_duplicate_cycles(cycles)

    def _init_dfs_state(self) -> dict:
        """Initialize DFS state for cycle detection."""
        white, gray, black = 0, 1, 2
        return {
            "white": white,
            "gray": gray,
            "black": black,
            "color": dict.fromkeys(self.rule_names, white),
            "path": [],
        }

    def _dfs_cycle_detection(self, node: str, state: dict, cycles: list[list[str]]) -> None:
        """Perform DFS for cycle detection."""
        state["color"][node] = state["gray"]
        state["path"].append(node)

        for neighbor in self.dependencies.get(node, set()):
            if neighbor in self.rule_names:  # Only check internal rules
                if state["color"][neighbor] == state["gray"]:
                    # Found cycle
                    cycle_start = state["path"].index(neighbor)
                    cycles.append(state["path"][cycle_start:])
                elif state["color"][neighbor] == state["white"]:
                    self._dfs_cycle_detection(neighbor, state, cycles)

        state["path"].pop()
        state["color"][node] = state["black"]

    def _remove_duplicate_cycles(self, cycles: list[list[str]]) -> list[list[str]]:
        """Remove duplicate cycles from the list."""
        unique_cycles = []
        for cycle in cycles:
            normalized = min(cycle[i:] + cycle[:i] for i in range(len(cycle)))
            if normalized not in unique_cycles:
                unique_cycles.append(normalized)

        return unique_cycles

    def _topological_sort(self) -> list[str] | None:
        """Perform topological sort on rules."""
        # Check for cycles first
        if self._find_circular_dependencies():
            return None

        in_degree = dict.fromkeys(self.rule_names, 0)

        # Calculate in-degrees: if rule A depends on rule B, then A has incoming edge from B
        for rule, deps in self.dependencies.items():
            in_degree[rule] = len([dep for dep in deps if dep in in_degree])

        # Find nodes with no incoming edges
        queue = [rule for rule, degree in in_degree.items() if degree == 0]
        result = []

        while queue:
            current = queue.pop(0)
            result.append(current)

            # Reduce in-degree for dependent nodes
            for rule, deps in self.dependencies.items():
                if current in deps and rule in in_degree:
                    in_degree[rule] -= 1
                    if in_degree[rule] == 0:
                        queue.append(rule)

        return result if len(result) == len(self.rule_names) else None

    # Visitor methods
    def visit_yara_file(self, node: YaraFile) -> None:
        for imp in node.imports:
            self.visit(imp)

        for inc in node.includes:
            self.visit(inc)

        for rule in node.rules:
            self.visit(rule)

    def visit_import(self, node: Import) -> None:
        self.imported_modules.add(node.module)

    def visit_include(self, node: Include) -> None:
        self.included_files.add(node.path)

    def visit_rule(self, node: Rule) -> None:
        self.current_rule = node.name

        # Check condition for rule references
        self.visit(node.condition)

        self.current_rule = None

    def visit_identifier(self, node: Identifier) -> None:
        # Check if identifier is a rule reference
        if self.current_rule and node.name in self.rule_names:
            self.dependencies[self.current_rule].add(node.name)

    def visit_function_call(self, node: FunctionCall) -> None:
        # Some functions might reference rules
        if node.function in self.rule_names and self.current_rule:
            self.dependencies[self.current_rule].add(node.function)

        # Visit arguments
        for arg in node.arguments:
            self.visit(arg)

    # Pass-through visitor methods
    def visit_binary_expression(self, node: BinaryExpression) -> None:
        self.visit(node.left)
        self.visit(node.right)

    def visit_unary_expression(self, node: UnaryExpression) -> None:
        self.visit(node.operand)

    def visit_parentheses_expression(self, node: ParenthesesExpression) -> None:
        self.visit(node.expression)

    def visit_for_expression(self, node: ForExpression) -> None:
        self.visit(node.iterable)
        self.visit(node.body)

    def visit_for_of_expression(self, node: ForOfExpression) -> None:
        self.visit(node.string_set)
        if node.condition:
            self.visit(node.condition)

    def visit_set_expression(self, node: SetExpression) -> None:
        for element in node.elements:
            self.visit(element)

    def visit_range_expression(self, node: RangeExpression) -> None:
        self.visit(node.low)
        self.visit(node.high)

    def visit_array_access(self, node: ArrayAccess) -> None:
        self.visit(node.array)
        self.visit(node.index)

    def visit_member_access(self, node: MemberAccess) -> None:
        self.visit(node.object)

    def visit_at_expression(self, node: AtExpression) -> None:
        self.visit(node.offset)

    def visit_in_expression(self, node: InExpression) -> None:
        self.visit(node.range)

    def visit_of_expression(self, node: OfExpression) -> None:
        self.visit(node.quantifier)
        self.visit(node.string_set)

    # No-op visitor methods
    def visit_tag(self, node: Tag) -> None:
        """Visit tag node - tags don't create dependencies."""

    def visit_string_definition(self, node: StringDefinition) -> None:
        """Visit string definition node - string definitions don't create dependencies."""

    def visit_plain_string(self, node: PlainString) -> None:
        """Visit plain string node - plain strings don't create dependencies."""

    def visit_hex_string(self, node: HexString) -> None:
        """Visit hex string node - hex strings don't create dependencies."""

    def visit_regex_string(self, node: RegexString) -> None:
        """Visit regex string node - regex strings don't create dependencies."""

    def visit_string_modifier(self, node: StringModifier) -> None:
        """Visit string modifier node - modifiers don't create dependencies."""

    def visit_hex_token(self, node: HexToken) -> None:
        """Visit hex token node - hex tokens don't create dependencies."""

    def visit_hex_byte(self, node: HexByte) -> None:
        """Visit hex byte node - hex bytes don't create dependencies."""

    def visit_hex_wildcard(self, node: HexWildcard) -> None:
        """Visit hex wildcard node - wildcards don't create dependencies."""

    def visit_hex_jump(self, node: HexJump) -> None:
        """Visit hex jump node - jumps don't create dependencies."""

    def visit_hex_alternative(self, node: HexAlternative) -> None:
        """Visit hex alternative node - alternatives don't create dependencies."""

    def visit_hex_nibble(self, node: HexNibble) -> None:
        """Visit hex nibble node - nibbles don't create dependencies."""

    def visit_expression(self, node: Expression) -> None:
        """Visit expression node - handled by specific expression type visitors."""

    def visit_string_identifier(self, node: StringIdentifier) -> None:
        """Visit string identifier node - string identifiers don't create rule dependencies."""

    def visit_string_wildcard(self, node: StringIdentifier) -> None:
        """Visit StringWildcard node."""
        pass

    def visit_string_count(self, node: StringCount) -> None:
        """Visit string count node - string counts don't create dependencies."""

    def visit_string_offset(self, node: StringOffset) -> None:
        """Visit string offset node - string offsets don't create dependencies."""

    def visit_string_length(self, node: StringLength) -> None:
        """Visit string length node - string lengths don't create dependencies."""

    def visit_integer_literal(self, node: IntegerLiteral) -> None:
        """Visit integer literal node - integer literals don't create dependencies."""

    def visit_double_literal(self, node: DoubleLiteral) -> None:
        """Visit double literal node - double literals don't create dependencies."""

    def visit_string_literal(self, node: StringLiteral) -> None:
        """Visit string literal node - string literals don't create dependencies."""

    def visit_boolean_literal(self, node: BooleanLiteral) -> None:
        """Visit boolean literal node - boolean literals don't create dependencies."""

    def visit_meta(self, node: Meta) -> None:
        """Visit meta node - meta fields don't create dependencies."""

    def visit_meta_statement(self, node) -> None:
        """Visit meta statement node - meta statements don't create dependencies."""

    def visit_condition(self, node) -> None:
        """Visit condition node - handled by specific condition visitors."""

    def visit_comment(self, node) -> None:
        """Visit comment node - comments don't create dependencies."""

    def visit_comment_group(self, node) -> None:
        """Visit comment group node - comment groups don't create dependencies."""

    def visit_module_reference(self, node) -> None:
        """Visit module reference node - module references are handled by imports."""

    def visit_dictionary_access(self, node) -> None:
        """Visit dictionary access node - dictionary access doesn't create rule dependencies."""

    def visit_defined_expression(self, node) -> None:
        """Visit defined expression node - defined expressions don't create rule dependencies."""

    def visit_string_operator_expression(self, node) -> None:
        """Visit string operator expression node - string operators don't create rule dependencies."""

    # Add missing abstract methods
    def visit_extern_import(self, node) -> None:
        """Visit extern import node - extern imports don't create rule dependencies."""

    def visit_extern_namespace(self, node) -> None:
        """Visit extern namespace node - extern namespaces don't create rule dependencies."""

    def visit_extern_rule(self, node) -> None:
        """Visit extern rule node - extern rules are handled separately from internal rules."""

    def visit_extern_rule_reference(self, node) -> None:
        """Visit extern rule reference node - extern rule references are handled separately."""

    def visit_in_rule_pragma(self, node) -> None:
        """Visit in-rule pragma node - pragmas don't create dependencies."""

    def visit_pragma(self, node) -> None:
        """Visit pragma node - pragmas don't create dependencies."""

    def visit_pragma_block(self, node) -> None:
        """Visit pragma block node - pragma blocks don't create dependencies."""

    def visit_regex_literal(self, node) -> None:
        """Visit regex literal node - regex literals don't create dependencies."""



"""YARA-L code generator from AST."""

from __future__ import annotations

from typing import TYPE_CHECKING

from yaraast.visitor import ASTVisitor

if TYPE_CHECKING:
    from yaraast.yaral.ast_nodes import (
        AggregationFunction,
        BinaryCondition,
        ConditionalExpression,
        ConditionExpression,
        ConditionSection,
        EventAssignment,
        EventCountCondition,
        EventExistsCondition,
        EventsSection,
        EventStatement,
        EventVariable,
        MatchSection,
        MatchVariable,
        MetaEntry,
        MetaSection,
        OptionsSection,
        OutcomeAssignment,
        OutcomeExpression,
        OutcomeSection,
        ReferenceList,
        RegexPattern,
        TimeWindow,
        UDMFieldAccess,
        UDMFieldPath,
        UnaryCondition,
        YaraLFile,
        YaraLRule,
    )


class YaraLGenerator(ASTVisitor[str]):
    """Generate YARA-L code from AST."""

    def __init__(self, indent_size: int = 2) -> None:
        """Initialize YARA-L generator.

        Args:
            indent_size: Number of spaces for indentation
        """
        self.indent_size = indent_size
        self.indent_level = 0

    def generate(self, ast: YaraLFile) -> str:
        """Generate YARA-L code from AST.

        Args:
            ast: YARA-L AST to generate code from

        Returns:
            Generated YARA-L code
        """
        return self.visit(ast)

    def _indent(self) -> str:
        """Get current indentation string."""
        return " " * (self.indent_level * self.indent_size)

    def _increase_indent(self) -> None:
        """Increase indentation level."""
        self.indent_level += 1

    def _decrease_indent(self) -> None:
        """Decrease indentation level."""
        self.indent_level = max(0, self.indent_level - 1)

    # File and Rule visitors
    def visit_yaral_file(self, node: YaraLFile) -> str:
        """Generate code for YARA-L file."""
        rules = []
        for rule in node.rules:
            rules.append(self.visit(rule))
        return "\n\n".join(rules)

    def visit_yaral_rule(self, node: YaraLRule) -> str:
        """Generate code for YARA-L rule."""
        parts = [f"rule {node.name} {{"]
        self._increase_indent()

        # Add meta section
        if node.meta:
            parts.append(self.visit(node.meta))

        # Add events section
        if node.events:
            parts.append(self.visit(node.events))

        # Add match section
        if node.match:
            parts.append(self.visit(node.match))

        # Add condition section
        if node.condition:
            parts.append(self.visit(node.condition))

        # Add outcome section
        if node.outcome:
            parts.append(self.visit(node.outcome))

        # Add options section
        if node.options:
            parts.append(self.visit(node.options))

        self._decrease_indent()
        parts.append("}")

        return "\n".join(filter(None, parts))

    # Section visitors
    def visit_meta_section(self, node: MetaSection) -> str:
        """Generate code for meta section."""
        if not node.entries:
            return ""

        lines = [f"{self._indent()}meta:"]
        self._increase_indent()

        for entry in node.entries:
            lines.append(self.visit(entry))

        self._decrease_indent()
        return "\n".join(lines)

    def visit_meta_entry(self, node: MetaEntry) -> str:
        """Generate code for meta entry."""
        value = node.value
        if isinstance(value, str):
            value = f'"{value}"'
        elif isinstance(value, bool):
            value = "true" if value else "false"

        return f"{self._indent()}{node.key} = {value}"

    def visit_events_section(self, node: EventsSection) -> str:
        """Generate code for events section."""
        if not node.statements:
            return ""

        lines = [f"{self._indent()}events:"]
        self._increase_indent()

        for statement in node.statements:
            lines.append(self.visit(statement))

        self._decrease_indent()
        return "\n".join(lines)

    def visit_event_statement(self, node: EventStatement) -> str:
        """Generate code for event statement."""
        parts = []

        # Add event variable
        if node.event:
            parts.append(self.visit(node.event))

        # Add assignments
        for assignment in node.assignments:
            parts.append(self.visit(assignment))

        # Join with AND if multiple parts
        if len(parts) > 1:
            return f"{self._indent()}{' and '.join(parts)}"
        if parts:
            return f"{self._indent()}{parts[0]}"
        return ""

    def visit_event_variable(self, node: EventVariable) -> str:
        """Generate code for event variable."""
        return node.name

    def visit_event_assignment(self, node: EventAssignment) -> str:
        """Generate code for event assignment."""
        field = self.visit(node.field_path)
        operator = node.operator

        # Handle value based on type
        value = node.value
        if isinstance(value, str):
            # Check if it's a reference or literal
            if not value.startswith("$") and not value.startswith("%"):
                value = f'"{value}"'
        elif hasattr(value, "accept"):
            # It's an AST node
            value = self.visit(value)

        return f"{field} {operator} {value}"

    def visit_udm_field_path(self, node: UDMFieldPath) -> str:
        """Generate code for UDM field path."""
        return ".".join(node.parts)

    def visit_udm_field_access(self, node: UDMFieldAccess) -> str:
        """Generate code for UDM field access."""
        event = self.visit(node.event)
        field = self.visit(node.field)
        return f"{event}.{field}"

    def visit_match_section(self, node: MatchSection) -> str:
        """Generate code for match section."""
        lines = [f"{self._indent()}match:"]
        self._increase_indent()

        # Add match variables
        for var in node.variables:
            lines.append(self.visit(var))

        # Add time window
        if node.time_window:
            lines.append(self.visit(node.time_window))

        self._decrease_indent()
        return "\n".join(lines)

    def visit_match_variable(self, node: MatchVariable) -> str:
        """Generate code for match variable."""
        line = f"{self._indent()}{node.name} = "

        if node.field:
            line += self.visit(node.field)

        if node.condition:
            line += f" over {node.condition}"

        return line

    def visit_time_window(self, node: TimeWindow) -> str:
        """Generate code for time window."""
        return f"{self._indent()}over {node.duration} {node.unit}"

    def visit_condition_section(self, node: ConditionSection) -> str:
        """Generate code for condition section."""
        lines = [f"{self._indent()}condition:"]
        self._increase_indent()

        lines.append(f"{self._indent()}{self.visit(node.expression)}")

        self._decrease_indent()
        return "\n".join(lines)

    def visit_condition_expression(self, node: ConditionExpression) -> str:
        """Generate code for condition expression."""
        # This is a base class, specific implementations below
        return ""

    def visit_binary_condition(self, node: BinaryCondition) -> str:
        """Generate code for binary condition."""
        left = self.visit(node.left) if hasattr(node, "left") and node.left else ""
        right = self.visit(node.right) if hasattr(node, "right") and node.right else ""
        operator = node.operator

        if operator in ["and", "or"]:
            return f"({left} {operator} {right})"
        return f"{left} {operator} {right}"

    def visit_unary_condition(self, node: UnaryCondition) -> str:
        """Generate code for unary condition."""
        operand = self.visit(node.operand) if hasattr(node, "operand") and node.operand else ""
        return f"{node.operator} {operand}"

    def visit_event_count_condition(self, node: EventCountCondition) -> str:
        """Generate code for event count condition."""
        return f"#{node.event} {node.operator} {node.count}"

    def visit_event_exists_condition(self, node: EventExistsCondition) -> str:
        """Generate code for event exists condition."""
        if node.negated:
            return f"not {node.event}"
        return node.event

    def visit_outcome_section(self, node: OutcomeSection) -> str:
        """Generate code for outcome section."""
        lines = [f"{self._indent()}outcome:"]
        self._increase_indent()

        # Add outcome variables
        for var in node.variables:
            lines.append(f"{self._indent()}{var} = {self.visit(node.variables[var])}")

        # Add conditional expressions
        for expr in node.conditional_expressions:
            lines.append(self.visit(expr))

        self._decrease_indent()
        return "\n".join(lines)

    def visit_outcome_assignment(self, node: OutcomeAssignment) -> str:
        """Generate code for outcome assignment."""
        return f"{node.variable} = {self.visit(node.expression)}"

    def visit_outcome_expression(self, node: OutcomeExpression) -> str:
        """Generate code for outcome expression."""
        if node.aggregation:
            return self.visit(node.aggregation)
        if node.field:
            return self.visit(node.field)
        if node.literal is not None:
            if isinstance(node.literal, str):
                return f'"{node.literal}"'
            return str(node.literal)
        return ""

    def visit_conditional_expression(self, node: ConditionalExpression) -> str:
        """Generate code for conditional expression."""
        condition = self.visit(node.condition)
        then_expr = self.visit(node.then_expression)
        else_expr = self.visit(node.else_expression) if node.else_expression else None

        if else_expr:
            return f"{self._indent()}if {condition} then {then_expr} else {else_expr}"
        return f"{self._indent()}if {condition} then {then_expr}"

    def visit_aggregation_function(self, node: AggregationFunction) -> str:
        """Generate code for aggregation function."""
        args = []
        for arg in node.arguments:
            if isinstance(arg, str):
                args.append(arg)
            else:
                args.append(self.visit(arg))

        args_str = ", ".join(args)
        return f"{node.function}({args_str})"

    def visit_reference_list(self, node: ReferenceList) -> str:
        """Generate code for reference list."""
        return f"%{node.name}"

    def visit_regex_pattern(self, node: RegexPattern) -> str:
        """Generate code for regex pattern."""
        modifiers = node.modifiers if node.modifiers else ""
        return f"/{node.pattern}/{modifiers}"

    def visit_options_section(self, node: OptionsSection) -> str:
        """Generate code for options section."""
        lines = [f"{self._indent()}options:"]
        self._increase_indent()

        for key, value in node.options.items():
            if isinstance(value, bool):
                value = "true" if value else "false"
            elif isinstance(value, str):
                value = f'"{value}"'
            lines.append(f"{self._indent()}{key} = {value}")

        self._decrease_indent()
        return "\n".join(lines)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: yara_ast.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    "",
    "yara_ast.proto",
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0eyara_ast.proto\x12\x07yaraast"\x93\x01\n\x08YaraFile\x12 \n\x07imports\x18\x01 \x03(\x0b\x32\x0f.yaraast.Import\x12"\n\x08includes\x18\x02 \x03(\x0b\x32\x10.yaraast.Include\x12\x1c\n\x05rules\x18\x03 \x03(\x0b\x32\r.yaraast.Rule\x12#\n\x08metadata\x18\x04 \x01(\x0b\x32\x11.yaraast.Metadata"\xa9\x01\n\x08Metadata\x12\x0e\n\x06\x66ormat\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x10\n\x08\x61st_type\x18\x03 \x01(\t\x12\x13\n\x0brules_count\x18\x04 \x01(\x05\x12\x15\n\rimports_count\x18\x05 \x01(\x05\x12\x16\n\x0eincludes_count\x18\x06 \x01(\x05\x12\x11\n\ttimestamp\x18\x07 \x01(\x03\x12\x13\n\x0bsource_file\x18\x08 \x01(\t"\'\n\x06Import\x12\x0e\n\x06module\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t"\x17\n\x07Include\x12\x0c\n\x04path\x18\x01 \x01(\t"\xff\x01\n\x04Rule\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tmodifiers\x18\x02 \x03(\t\x12\x1a\n\x04tags\x18\x03 \x03(\x0b\x32\x0c.yaraast.Tag\x12%\n\x04meta\x18\x04 \x03(\x0b\x32\x17.yaraast.Rule.MetaEntry\x12*\n\x07strings\x18\x05 \x03(\x0b\x32\x19.yaraast.StringDefinition\x12&\n\tcondition\x18\x06 \x01(\x0b\x32\x13.yaraast.Expression\x1a?\n\tMetaEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.yaraast.MetaValue:\x02\x38\x01"\x13\n\x03Tag\x12\x0c\n\x04name\x18\x01 \x01(\t"o\n\tMetaValue\x12\x16\n\x0cstring_value\x18\x01 \x01(\tH\x00\x12\x13\n\tint_value\x18\x02 \x01(\x03H\x00\x12\x14\n\nbool_value\x18\x03 \x01(\x08H\x00\x12\x16\n\x0c\x64ouble_value\x18\x04 \x01(\x01H\x00\x42\x07\n\x05value"\xa6\x01\n\x10StringDefinition\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12%\n\x05plain\x18\x02 \x01(\x0b\x32\x14.yaraast.PlainStringH\x00\x12!\n\x03hex\x18\x03 \x01(\x0b\x32\x12.yaraast.HexStringH\x00\x12%\n\x05regex\x18\x04 \x01(\x0b\x32\x14.yaraast.RegexStringH\x00\x42\r\n\x0bstring_type"H\n\x0bPlainString\x12\r\n\x05value\x18\x01 \x01(\t\x12*\n\tmodifiers\x18\x02 \x03(\x0b\x32\x17.yaraast.StringModifier"Z\n\tHexString\x12!\n\x06tokens\x18\x01 \x03(\x0b\x32\x11.yaraast.HexToken\x12*\n\tmodifiers\x18\x02 \x03(\x0b\x32\x17.yaraast.StringModifier"H\n\x0bRegexString\x12\r\n\x05regex\x18\x01 \x01(\t\x12*\n\tmodifiers\x18\x02 \x03(\x0b\x32\x17.yaraast.StringModifier"-\n\x0eStringModifier\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t"\xdc\x01\n\x08HexToken\x12 \n\x04\x62yte\x18\x01 \x01(\x0b\x32\x10.yaraast.HexByteH\x00\x12(\n\x08wildcard\x18\x02 \x01(\x0b\x32\x14.yaraast.HexWildcardH\x00\x12 \n\x04jump\x18\x03 \x01(\x0b\x32\x10.yaraast.HexJumpH\x00\x12.\n\x0b\x61lternative\x18\x04 \x01(\x0b\x32\x17.yaraast.HexAlternativeH\x00\x12$\n\x06nibble\x18\x05 \x01(\x0b\x32\x12.yaraast.HexNibbleH\x00\x42\x0c\n\ntoken_type"\x18\n\x07HexByte\x12\r\n\x05value\x18\x01 \x01(\t"\r\n\x0bHexWildcard"-\n\x07HexJump\x12\x10\n\x08min_jump\x18\x01 \x01(\x05\x12\x10\n\x08max_jump\x18\x02 \x01(\x05"=\n\x0eHexAlternative\x12+\n\x0c\x61lternatives\x18\x01 \x03(\x0b\x32\x15.yaraast.HexTokenList"1\n\x0cHexTokenList\x12!\n\x06tokens\x18\x01 \x03(\x0b\x32\x11.yaraast.HexToken"(\n\tHexNibble\x12\x0c\n\x04high\x18\x01 \x01(\x08\x12\r\n\x05value\x18\x02 \x01(\x05"\xae\n\n\nExpression\x12)\n\nidentifier\x18\x01 \x01(\x0b\x32\x13.yaraast.IdentifierH\x00\x12\x36\n\x11string_identifier\x18\x02 \x01(\x0b\x32\x19.yaraast.StringIdentifierH\x00\x12,\n\x0cstring_count\x18\x03 \x01(\x0b\x32\x14.yaraast.StringCountH\x00\x12.\n\rstring_offset\x18\x04 \x01(\x0b\x32\x15.yaraast.StringOffsetH\x00\x12.\n\rstring_length\x18\x05 \x01(\x0b\x32\x15.yaraast.StringLengthH\x00\x12\x32\n\x0finteger_literal\x18\x06 \x01(\x0b\x32\x17.yaraast.IntegerLiteralH\x00\x12\x30\n\x0e\x64ouble_literal\x18\x07 \x01(\x0b\x32\x16.yaraast.DoubleLiteralH\x00\x12\x30\n\x0estring_literal\x18\x08 \x01(\x0b\x32\x16.yaraast.StringLiteralH\x00\x12.\n\rregex_literal\x18\t \x01(\x0b\x32\x15.yaraast.RegexLiteralH\x00\x12\x32\n\x0f\x62oolean_literal\x18\n \x01(\x0b\x32\x17.yaraast.BooleanLiteralH\x00\x12\x36\n\x11\x62inary_expression\x18\x0b \x01(\x0b\x32\x19.yaraast.BinaryExpressionH\x00\x12\x34\n\x10unary_expression\x18\x0c \x01(\x0b\x32\x18.yaraast.UnaryExpressionH\x00\x12@\n\x16parentheses_expression\x18\r \x01(\x0b\x32\x1e.yaraast.ParenthesesExpressionH\x00\x12\x30\n\x0eset_expression\x18\x0e \x01(\x0b\x32\x16.yaraast.SetExpressionH\x00\x12\x34\n\x10range_expression\x18\x0f \x01(\x0b\x32\x18.yaraast.RangeExpressionH\x00\x12.\n\rfunction_call\x18\x10 \x01(\x0b\x32\x15.yaraast.FunctionCallH\x00\x12,\n\x0c\x61rray_access\x18\x11 \x01(\x0b\x32\x14.yaraast.ArrayAccessH\x00\x12.\n\rmember_access\x18\x12 \x01(\x0b\x32\x15.yaraast.MemberAccessH\x00\x12\x30\n\x0e\x66or_expression\x18\x13 \x01(\x0b\x32\x16.yaraast.ForExpressionH\x00\x12\x35\n\x11\x66or_of_expression\x18\x14 \x01(\x0b\x32\x18.yaraast.ForOfExpressionH\x00\x12.\n\rat_expression\x18\x15 \x01(\x0b\x32\x15.yaraast.AtExpressionH\x00\x12.\n\rin_expression\x18\x16 \x01(\x0b\x32\x15.yaraast.InExpressionH\x00\x12.\n\rof_expression\x18\x17 \x01(\x0b\x32\x15.yaraast.OfExpressionH\x00\x12\x38\n\x12\x64\x65\x66ined_expression\x18\x18 \x01(\x0b\x32\x1a.yaraast.DefinedExpressionH\x00\x12G\n\x1astring_operator_expression\x18\x19 \x01(\x0b\x32!.yaraast.StringOperatorExpressionH\x00\x42\x11\n\x0f\x65xpression_type"\x1a\n\nIdentifier\x12\x0c\n\x04name\x18\x01 \x01(\t" \n\x10StringIdentifier\x12\x0c\n\x04name\x18\x01 \x01(\t" \n\x0bStringCount\x12\x11\n\tstring_id\x18\x01 \x01(\t"E\n\x0cStringOffset\x12\x11\n\tstring_id\x18\x01 \x01(\t\x12"\n\x05index\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression"E\n\x0cStringLength\x12\x11\n\tstring_id\x18\x01 \x01(\t\x12"\n\x05index\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression"\x1f\n\x0eIntegerLiteral\x12\r\n\x05value\x18\x01 \x01(\x03"\x1e\n\rDoubleLiteral\x12\r\n\x05value\x18\x01 \x01(\x01"\x1e\n\rStringLiteral\x12\r\n\x05value\x18\x01 \x01(\t"2\n\x0cRegexLiteral\x12\x0f\n\x07pattern\x18\x01 \x01(\t\x12\x11\n\tmodifiers\x18\x02 \x01(\t"\x1f\n\x0e\x42ooleanLiteral\x12\r\n\x05value\x18\x01 \x01(\x08"k\n\x10\x42inaryExpression\x12!\n\x04left\x18\x01 \x01(\x0b\x32\x13.yaraast.Expression\x12\x10\n\x08operator\x18\x02 \x01(\t\x12"\n\x05right\x18\x03 \x01(\x0b\x32\x13.yaraast.Expression"I\n\x0fUnaryExpression\x12\x10\n\x08operator\x18\x01 \x01(\t\x12$\n\x07operand\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression"@\n\x15ParenthesesExpression\x12\'\n\nexpression\x18\x01 \x01(\x0b\x32\x13.yaraast.Expression"6\n\rSetExpression\x12%\n\x08\x65lements\x18\x01 \x03(\x0b\x32\x13.yaraast.Expression"V\n\x0fRangeExpression\x12 \n\x03low\x18\x01 \x01(\x0b\x32\x13.yaraast.Expression\x12!\n\x04high\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression"H\n\x0c\x46unctionCall\x12\x10\n\x08\x66unction\x18\x01 \x01(\t\x12&\n\targuments\x18\x02 \x03(\x0b\x32\x13.yaraast.Expression"U\n\x0b\x41rrayAccess\x12"\n\x05\x61rray\x18\x01 \x01(\x0b\x32\x13.yaraast.Expression\x12"\n\x05index\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression"C\n\x0cMemberAccess\x12#\n\x06object\x18\x01 \x01(\x0b\x32\x13.yaraast.Expression\x12\x0e\n\x06member\x18\x02 \x01(\t"\x7f\n\rForExpression\x12\x12\n\nquantifier\x18\x01 \x01(\t\x12\x10\n\x08variable\x18\x02 \x01(\t\x12%\n\x08iterable\x18\x03 \x01(\x0b\x32\x13.yaraast.Expression\x12!\n\x04\x62ody\x18\x04 \x01(\x0b\x32\x13.yaraast.Expression"v\n\x0f\x46orOfExpression\x12\x12\n\nquantifier\x18\x01 \x01(\t\x12\'\n\nstring_set\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression\x12&\n\tcondition\x18\x03 \x01(\x0b\x32\x13.yaraast.Expression"F\n\x0c\x41tExpression\x12\x11\n\tstring_id\x18\x01 \x01(\t\x12#\n\x06offset\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression"E\n\x0cInExpression\x12\x11\n\tstring_id\x18\x01 \x01(\t\x12"\n\x05range\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression"`\n\x0cOfExpression\x12\'\n\nquantifier\x18\x01 \x01(\x0b\x32\x13.yaraast.Expression\x12\'\n\nstring_set\x18\x02 \x01(\x0b\x32\x13.yaraast.Expression"<\n\x11\x44\x65\x66inedExpression\x12\'\n\nexpression\x18\x01 \x01(\x0b\x32\x13.yaraast.Expression"s\n\x18StringOperatorExpression\x12!\n\x04left\x18\x01 \x01(\x0b\x32\x13.yaraast.Expression\x12\x10\n\x08operator\x18\x02 \x01(\t\x12"\n\x05right\x18\x03 \x01(\x0b\x32\x13.yaraast.Expressionb\x06proto3',
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "yara_ast_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_RULE_METAENTRY"]._loaded_options = None
    _globals["_RULE_METAENTRY"]._serialized_options = b"8\001"
    _globals["_YARAFILE"]._serialized_start = 28
    _globals["_YARAFILE"]._serialized_end = 175
    _globals["_METADATA"]._serialized_start = 178
    _globals["_METADATA"]._serialized_end = 347
    _globals["_IMPORT"]._serialized_start = 349
    _globals["_IMPORT"]._serialized_end = 388
    _globals["_INCLUDE"]._serialized_start = 390
    _globals["_INCLUDE"]._serialized_end = 413
    _globals["_RULE"]._serialized_start = 416
    _globals["_RULE"]._serialized_end = 671
    _globals["_RULE_METAENTRY"]._serialized_start = 608
    _globals["_RULE_METAENTRY"]._serialized_end = 671
    _globals["_TAG"]._serialized_start = 673
    _globals["_TAG"]._serialized_end = 692
    _globals["_METAVALUE"]._serialized_start = 694
    _globals["_METAVALUE"]._serialized_end = 805
    _globals["_STRINGDEFINITION"]._serialized_start = 808
    _globals["_STRINGDEFINITION"]._serialized_end = 974
    _globals["_PLAINSTRING"]._serialized_start = 976
    _globals["_PLAINSTRING"]._serialized_end = 1048
    _globals["_HEXSTRING"]._serialized_start = 1050
    _globals["_HEXSTRING"]._serialized_end = 1140
    _globals["_REGEXSTRING"]._serialized_start = 1142
    _globals["_REGEXSTRING"]._serialized_end = 1214
    _globals["_STRINGMODIFIER"]._serialized_start = 1216
    _globals["_STRINGMODIFIER"]._serialized_end = 1261
    _globals["_HEXTOKEN"]._serialized_start = 1264
    _globals["_HEXTOKEN"]._serialized_end = 1484
    _globals["_HEXBYTE"]._serialized_start = 1486
    _globals["_HEXBYTE"]._serialized_end = 1510
    _globals["_HEXWILDCARD"]._serialized_start = 1512
    _globals["_HEXWILDCARD"]._serialized_end = 1525
    _globals["_HEXJUMP"]._serialized_start = 1527
    _globals["_HEXJUMP"]._serialized_end = 1572
    _globals["_HEXALTERNATIVE"]._serialized_start = 1574
    _globals["_HEXALTERNATIVE"]._serialized_end = 1635
    _globals["_HEXTOKENLIST"]._serialized_start = 1637
    _globals["_HEXTOKENLIST"]._serialized_end = 1686
    _globals["_HEXNIBBLE"]._serialized_start = 1688
    _globals["_HEXNIBBLE"]._serialized_end = 1728
    _globals["_EXPRESSION"]._serialized_start = 1731
    _globals["_EXPRESSION"]._serialized_end = 3057
    _globals["_IDENTIFIER"]._serialized_start = 3059
    _globals["_IDENTIFIER"]._serialized_end = 3085
    _globals["_STRINGIDENTIFIER"]._serialized_start = 3087
    _globals["_STRINGIDENTIFIER"]._serialized_end = 3119
    _globals["_STRINGCOUNT"]._serialized_start = 3121
    _globals["_STRINGCOUNT"]._serialized_end = 3153
    _globals["_STRINGOFFSET"]._serialized_start = 3155
    _globals["_STRINGOFFSET"]._serialized_end = 3224
    _globals["_STRINGLENGTH"]._serialized_start = 3226
    _globals["_STRINGLENGTH"]._serialized_end = 3295
    _globals["_INTEGERLITERAL"]._serialized_start = 3297
    _globals["_INTEGERLITERAL"]._serialized_end = 3328
    _globals["_DOUBLELITERAL"]._serialized_start = 3330
    _globals["_DOUBLELITERAL"]._serialized_end = 3360
    _globals["_STRINGLITERAL"]._serialized_start = 3362
    _globals["_STRINGLITERAL"]._serialized_end = 3392
    _globals["_REGEXLITERAL"]._serialized_start = 3394
    _globals["_REGEXLITERAL"]._serialized_end = 3444
    _globals["_BOOLEANLITERAL"]._serialized_start = 3446
    _globals["_BOOLEANLITERAL"]._serialized_end = 3477
    _globals["_BINARYEXPRESSION"]._serialized_start = 3479
    _globals["_BINARYEXPRESSION"]._serialized_end = 3586
    _globals["_UNARYEXPRESSION"]._serialized_start = 3588
    _globals["_UNARYEXPRESSION"]._serialized_end = 3661
    _globals["_PARENTHESESEXPRESSION"]._serialized_start = 3663
    _globals["_PARENTHESESEXPRESSION"]._serialized_end = 3727
    _globals["_SETEXPRESSION"]._serialized_start = 3729
    _globals["_SETEXPRESSION"]._serialized_end = 3783
    _globals["_RANGEEXPRESSION"]._serialized_start = 3785
    _globals["_RANGEEXPRESSION"]._serialized_end = 3871
    _globals["_FUNCTIONCALL"]._serialized_start = 3873
    _globals["_FUNCTIONCALL"]._serialized_end = 3945
    _globals["_ARRAYACCESS"]._serialized_start = 3947
    _globals["_ARRAYACCESS"]._serialized_end = 4032
    _globals["_MEMBERACCESS"]._serialized_start = 4034
    _globals["_MEMBERACCESS"]._serialized_end = 4101
    _globals["_FOREXPRESSION"]._serialized_start = 4103
    _globals["_FOREXPRESSION"]._serialized_end = 4230
    _globals["_FOROFEXPRESSION"]._serialized_start = 4232
    _globals["_FOROFEXPRESSION"]._serialized_end = 4350
    _globals["_ATEXPRESSION"]._serialized_start = 4352
    _globals["_ATEXPRESSION"]._serialized_end = 4422
    _globals["_INEXPRESSION"]._serialized_start = 4424
    _globals["_INEXPRESSION"]._serialized_end = 4493
    _globals["_OFEXPRESSION"]._serialized_start = 4495
    _globals["_OFEXPRESSION"]._serialized_end = 4591
    _globals["_DEFINEDEXPRESSION"]._serialized_start = 4593
    _globals["_DEFINEDEXPRESSION"]._serialized_end = 4653
    _globals["_STRINGOPERATOREXPRESSION"]._serialized_start = 4655
    _globals["_STRINGOPERATOREXPRESSION"]._serialized_end = 4770
# @@protoc_insertion_point(module_scope)



"""CLI commands for AST-based analysis."""

import sys
from pathlib import Path

import click
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

from yaraast.analysis import BestPracticesAnalyzer, OptimizationAnalyzer
from yaraast.parser import Parser

console = Console()


@click.group()
def analyze() -> None:
    """AST-based analysis commands."""


@analyze.command()
@click.argument("rule_file", type=click.Path(exists=True))
@click.option("-v", "--verbose", is_flag=True, help="Show all suggestions")
@click.option(
    "-c",
    "--category",
    type=click.Choice(["style", "optimization", "structure", "all"]),
    default="all",
    help="Filter by category",
)
def best_practices(rule_file: str, verbose: bool, category: str) -> None:
    """Analyze YARA rules for best practices using AST.

    This is not a full linter but an AST-based analyzer that identifies
    patterns and suggests improvements based on the rule structure.

    Example:
        yaraast analyze best-practices rules.yar
        yaraast analyze best-practices rules.yar -c style

    """
    try:
        ast = _parse_rule_file(rule_file)
        report = _analyze_best_practices(ast)
        _display_best_practices_report(rule_file, report, verbose, category)
    except Exception as e:
        console.print(f"[red]Error:[/red] {e}", style="bold")
        sys.exit(1)


def _parse_rule_file(rule_file: str):
    """Parse a YARA rule file."""
    with Path(rule_file).open() as f:
        content = f.read()
    parser = Parser()
    return parser.parse(content)


def _analyze_best_practices(ast):
    """Analyze AST for best practices."""
    analyzer = BestPracticesAnalyzer()
    return analyzer.analyze(ast)


def _display_best_practices_report(
    rule_file: str,
    report,
    verbose: bool,
    category: str,
) -> None:
    """Display best practices report."""
    console.print(f"\n[bold]Best Practices Analysis:[/bold] {rule_file}\n")

    errors, warnings, info = _get_severity_counts(report)
    _display_summary(errors, warnings, info)

    suggestions = _filter_suggestions(report.suggestions, category)
    _display_issues(suggestions)

    if verbose:
        _display_verbose_info(suggestions, report)

    _handle_exit_code(errors, warnings, info, verbose)


def _get_severity_counts(report):
    """Get counts by severity."""
    errors = report.get_by_severity("error")
    warnings = report.get_by_severity("warning")
    info = report.get_by_severity("info")
    return errors, warnings, info


def _display_summary(errors, warnings, info) -> None:
    """Display summary table."""
    summary = Table(show_header=False, box=None)
    summary.add_row("✗ Errors:", f"[red]{len(errors)}[/red]")
    summary.add_row("⚠ Warnings:", f"[yellow]{len(warnings)}[/yellow]")
    summary.add_row("i Info:", f"[blue]{len(info)}[/blue]")
    console.print(Panel(summary, title="Summary", width=30))


def _filter_suggestions(suggestions, category):
    """Filter suggestions by category."""
    if category != "all":
        return [s for s in suggestions if s.category == category]
    return suggestions


def _display_issues(suggestions) -> None:
    """Display errors and warnings."""
    important = [s for s in suggestions if s.severity in ("error", "warning")]
    if important:
        console.print("\n[bold red]Issues:[/bold red]")
        for suggestion in important:
            console.print(f"  {suggestion.format()}")


def _display_verbose_info(suggestions, report) -> None:
    """Display verbose information."""
    info_items = [s for s in suggestions if s.severity == "info"]
    if info_items:
        console.print("\n[bold blue]Suggestions:[/bold blue]")
        for suggestion in info_items:
            console.print(f"  {suggestion.format()}")

    if report.statistics:
        console.print("\n[dim]Statistics:[/dim]")
        for key, value in report.statistics.items():
            console.print(f"  {key}: {value}")


def _handle_exit_code(errors, warnings, info, verbose) -> None:
    """Handle exit code based on results."""
    if errors:
        sys.exit(2)
    elif warnings and not verbose:
        console.print(f"\n[dim]Use -v to see {len(info)} additional suggestions[/dim]")
        sys.exit(1)
    else:
        console.print("\n[green]✓ No major issues found[/green]")
        sys.exit(0)


def _optimize_display_impact_summary(report) -> None:
    """Display impact summary table."""
    table = Table(title="Optimization Opportunities")
    table.add_column("Impact", style="cyan")
    table.add_column("Count", justify="right")

    # Ensure statistics are populated
    if not report.statistics:
        report.statistics = {}
    if "by_impact" not in report.statistics:
        report.statistics["by_impact"] = {
            "high": sum(1 for s in report.suggestions if s.impact == "high"),
            "medium": sum(1 for s in report.suggestions if s.impact == "medium"),
            "low": sum(1 for s in report.suggestions if s.impact == "low"),
        }

    for level in ["high", "medium", "low"]:
        count = report.statistics["by_impact"].get(level, 0)
        style = {"high": "red", "medium": "yellow", "low": "green"}.get(level, "white")
        table.add_row(f"[{style}]{level.capitalize()}[/{style}]", str(count))

    console.print(table)


def _optimize_display_suggestions(suggestions, verbose) -> None:
    """Display optimization suggestions by impact level."""
    for level in ["high", "medium", "low"]:
        level_suggestions = [s for s in suggestions if s.impact == level]
        if level_suggestions:
            _display_suggestions_by_level(level_suggestions, level, verbose)


def _display_suggestions_by_level(level_suggestions, level, verbose) -> None:
    """Display suggestions for a specific impact level."""
    style = _get_level_style(level)
    console.print(f"\n[bold {style}]{level.capitalize()} Impact:[/bold {style}]")

    for suggestion in level_suggestions:
        console.print(f"  {suggestion.format()}")
        if verbose:
            _display_code_examples(suggestion)


def _get_level_style(level: str) -> str:
    """Get console style for impact level."""
    return {"high": "red", "medium": "yellow", "low": "blue"}.get(level, "white")


def _display_code_examples(suggestion) -> None:
    """Display before/after code examples if available."""
    if suggestion.code_before or suggestion.code_after:
        if suggestion.code_before:
            console.print(f"    Before: [dim]{suggestion.code_before}[/dim]")
        if suggestion.code_after:
            console.print(f"    After:  [green]{suggestion.code_after}[/green]")


@analyze.command()
@click.argument("rule_file", type=click.Path(exists=True))
@click.option("-v", "--verbose", is_flag=True, help="Show all optimizations")
@click.option(
    "-i",
    "--impact",
    type=click.Choice(["low", "medium", "high", "all"]),
    default="all",
    help="Filter by impact level",
)
def optimize(rule_file: str, verbose: bool, impact: str) -> None:
    """Analyze YARA rules for optimization opportunities.

    Uses AST analysis to identify patterns that could be optimized,
    including redundant strings, complex conditions, and more.

    Example:
        yaraast analyze optimize rules.yar
        yaraast analyze optimize rules.yar -i high

    """
    try:
        ast = _parse_rule_file(rule_file)
        analyzer = OptimizationAnalyzer()
        report = analyzer.analyze(ast)

        console.print(f"\n[bold]Optimization Analysis:[/bold] {rule_file}\n")
        _optimize_display_impact_summary(report)

        suggestions = report.suggestions
        if impact != "all":
            suggestions = [s for s in suggestions if s.impact == impact]

        _optimize_display_suggestions(suggestions, verbose)

        if verbose and report.statistics:
            console.print("\n[dim]Analysis Statistics:[/dim]")
            console.print(
                f"  Total suggestions: {report.statistics['total_suggestions']}",
            )

        # Exit based on results
        if report.high_impact_count > 0:
            console.print(
                f"\n[yellow]⚠ Found {report.high_impact_count} high-impact optimization opportunities[/yellow]",
            )
            sys.exit(1)
        elif len(report.suggestions) > 0:
            console.print(
                f"\n[blue]i Found {len(report.suggestions)} optimization suggestions[/blue]",
            )
            sys.exit(0)
        else:
            console.print("\n[green]✓ No optimization opportunities found[/green]")
            sys.exit(0)
    except Exception as e:
        console.print(f"[red]Error:[/red] {e}", style="bold")
        sys.exit(1)


@analyze.command()
@click.argument("rule_file", type=click.Path(exists=True))
@click.option("-o", "--output", type=click.Path(), help="Save report to file")
@click.option(
    "-f",
    "--format",
    type=click.Choice(["text", "json"]),
    default="text",
    help="Output format",
)
def _generate_json_report(rule_file, bp_report, opt_report):
    """Generate JSON format report."""
    return {
        "file": rule_file,
        "best_practices": {
            "suggestions": [
                {
                    "rule": s.rule_name,
                    "category": s.category,
                    "severity": s.severity,
                    "message": s.message,
                    "location": s.location,
                }
                for s in bp_report.suggestions
            ],
            "statistics": bp_report.statistics,
        },
        "optimizations": {
            "suggestions": [
                {
                    "rule": s.rule_name,
                    "type": s.optimization_type,
                    "impact": s.impact,
                    "description": s.description,
                }
                for s in opt_report.suggestions
            ],
            "statistics": opt_report.statistics,
        },
    }


def _generate_text_report(rule_file, bp_report, opt_report):
    """Generate text format report."""
    lines = []
    lines.append(f"AST Analysis Report: {rule_file}")
    lines.append("=" * 50)

    _add_best_practices_section(lines, bp_report)
    _add_optimizations_section(lines, opt_report)
    _add_summary_section(lines, bp_report, opt_report)

    return "\n".join(lines)


def _add_best_practices_section(lines, bp_report) -> None:
    """Add best practices section to report."""
    lines.append("\nBEST PRACTICES")
    lines.append("-" * 20)

    for severity in ["error", "warning", "info"]:
        items = bp_report.get_by_severity(severity)
        if items:
            lines.append(f"\n{severity.upper()}S ({len(items)}):")
            for s in items:
                lines.append(f"  {s.format()}")


def _add_optimizations_section(lines, opt_report) -> None:
    """Add optimizations section to report."""
    lines.append("\n\nOPTIMIZATIONS")
    lines.append("-" * 20)

    for impact_level in ["high", "medium", "low"]:
        items = [s for s in opt_report.suggestions if s.impact == impact_level]
        if items:
            lines.append(f"\n{impact_level.upper()} IMPACT ({len(items)}):")
            for s in items:
                lines.append(f"  {s.format()}")


def _add_summary_section(lines, bp_report, opt_report) -> None:
    """Add summary section to report."""
    lines.append("\n\nSUMMARY")
    lines.append("-" * 20)
    lines.append(
        f"Total issues: {len(bp_report.get_by_severity('error')) + len(bp_report.get_by_severity('warning'))}",
    )
    lines.append(
        f"Total suggestions: {len(bp_report.suggestions) + len(opt_report.suggestions)}",
    )


def full(rule_file: str, output: str, format: str) -> None:
    """Run full AST-based analysis (best practices + optimizations).

    Combines both best practices checking and optimization analysis
    into a comprehensive report.

    Example:
        yaraast analyze full rules.yar
        yaraast analyze full rules.yar -o report.txt

    """
    try:
        ast = _parse_rule_file(rule_file)

        bp_analyzer = BestPracticesAnalyzer()
        bp_report = bp_analyzer.analyze(ast)

        opt_analyzer = OptimizationAnalyzer()
        opt_report = opt_analyzer.analyze(ast)

        if format == "json":
            import json

            report_data = _generate_json_report(rule_file, bp_report, opt_report)

            if output:
                with Path(output).open("w") as f:
                    json.dump(report_data, f, indent=2)
                console.print(f"[green]Report saved to {output}[/green]")
            else:
                console.print(json.dumps(report_data, indent=2))
        else:
            report_text = _generate_text_report(rule_file, bp_report, opt_report)

            if output:
                with Path(output).open("w") as f:
                    f.write(report_text)
                console.print(f"[green]Report saved to {output}[/green]")
            else:
                console.print(report_text)

        # Exit code based on issues
        has_errors = len(bp_report.get_by_severity("error")) > 0
        has_high_impact = opt_report.high_impact_count > 0

        if has_errors:
            sys.exit(2)
        elif has_high_impact:
            sys.exit(1)
        else:
            sys.exit(0)
    except Exception as e:
        console.print(f"[red]Error:[/red] {e}", style="bold")
        sys.exit(1)



"""Multi-file resolution package."""

from .dependency_graph import DependencyGraph, DependencyNode
from .include_resolver import IncludeResolver
from .workspace import Workspace, WorkspaceAnalyzer

__all__ = [
    "DependencyGraph",
    "DependencyNode",
    "IncludeResolver",
    "Workspace",
    "WorkspaceAnalyzer",
]



"""Memory optimization utilities for YARA AST processing."""

from __future__ import annotations

import gc
import weakref
from typing import TYPE_CHECKING, Any

from yaraast.visitor.visitor import ASTTransformer

if TYPE_CHECKING:
    from yaraast.ast.base import ASTNode, YaraFile
    from yaraast.ast.rules import Rule


class MemoryOptimizer:
    """Optimizes memory usage when processing large YARA ASTs."""

    def __init__(
        self,
        aggressive: bool = False,
        memory_limit_mb: int | None = None,
        gc_threshold: int | None = None,
        enable_tracking: bool = False,
    ) -> None:
        """Initialize memory optimizer.

        Args:
            aggressive: If True, applies more aggressive memory optimizations
            memory_limit_mb: Memory limit in MB
            gc_threshold: Garbage collection threshold
            enable_tracking: Enable object tracking

        """
        self.aggressive = aggressive
        self.memory_limit_mb = memory_limit_mb
        self.gc_threshold = gc_threshold or 10
        self.enable_tracking = enable_tracking
        self._cache = weakref.WeakValueDictionary()
        self._string_pool = {}
        self._tracked_objects: list[Any] = []
        self._ast_pool: list[Any] = []
        self._stats = {
            "nodes_processed": 0,
            "strings_pooled": 0,
            "memory_saved": 0,
            "total_objects": 0,
        }

    def optimize(self, yara_file: YaraFile) -> YaraFile:
        """Optimize memory usage for a YARA file."""
        # Clear caches
        self._string_pool.clear()

        # Optimize the AST
        optimizer = MemoryOptimizerTransformer(self._string_pool, self.aggressive)
        _ = optimizer.visit(yara_file)

        # Update stats
        self._stats["nodes_processed"] += optimizer.nodes_processed
        self._stats["strings_pooled"] += len(self._string_pool)

        # Force garbage collection if aggressive
        if self.aggressive:
            gc.collect()

        return yara_file

    def optimize_rule(self, rule: Rule) -> Rule:
        """Optimize memory usage for a single rule."""
        optimizer = MemoryOptimizerTransformer(self._string_pool, self.aggressive)
        return optimizer.visit(rule)

    def optimize_rules(self, rules: list[Rule]) -> list[Rule]:
        """Optimize memory usage for a list of rules."""
        return [self.optimize_rule(rule) for rule in rules]

    def get_memory_usage(self) -> dict[str, Any]:
        """Get current memory usage statistics."""
        import os

        import psutil

        process = psutil.Process(os.getpid())
        mem_info = process.memory_info()

        return {
            "rss_mb": mem_info.rss / 1024 / 1024,
            "vms_mb": mem_info.vms / 1024 / 1024,
            "percent": process.memory_percent(),
            "available_mb": psutil.virtual_memory().available / 1024 / 1024,
        }

    def clear_caches(self) -> None:
        """Clear all internal caches."""
        self._cache.clear()
        self._string_pool.clear()
        gc.collect()

    def get_statistics(self) -> dict[str, Any]:
        """Get optimization statistics."""
        return {
            **self._stats,
            "string_pool_size": len(self._string_pool),
            "cache_size": len(self._cache),
        }

    def memory_managed_context(self):
        """Create a memory-managed context."""
        from contextlib import contextmanager

        @contextmanager
        def context():
            try:
                yield
            finally:
                # Cleanup tracked objects
                if self.enable_tracking:
                    self._tracked_objects.clear()
                gc.collect()

        return context()

    def track_object(self, obj: Any) -> None:
        """Track an object for memory management."""
        if self.enable_tracking:
            self._tracked_objects.append(obj)
            self._stats["total_objects"] = len(self._tracked_objects)

    def get_memory_stats(self):
        """Get memory statistics as an object."""
        from dataclasses import dataclass

        @dataclass
        class MemoryStats:
            total_objects: int
            nodes_processed: int
            strings_pooled: int

        return MemoryStats(
            total_objects=self._stats.get("total_objects", 0),
            nodes_processed=self._stats.get("nodes_processed", 0),
            strings_pooled=self._stats.get("strings_pooled", 0),
        )

    def force_cleanup(self) -> int:
        """Force garbage collection and cleanup."""
        # Clear weak references
        self._tracked_objects.clear()
        self._stats["total_objects"] = 0
        # Run garbage collection
        return gc.collect()

    def create_memory_efficient_ast(self):
        """Create or reuse an AST from pool."""
        from yaraast.ast.base import YaraFile

        # Try to reuse from pool
        if self._ast_pool:
            return self._ast_pool.pop()

        # Create new minimal AST
        return YaraFile(imports=[], includes=[], rules=[])

    def return_ast_to_pool(self, ast: Any) -> None:
        """Return an AST to the pool for reuse."""
        self._ast_pool.append(ast)

    def batch_process_with_memory_limit(
        self,
        items: list[Any],
        processor: Any,
        batch_size: int = 10,
    ):
        """Process items in batches with memory management."""
        # Split into batches
        for i in range(0, len(items), batch_size):
            batch = items[i : i + batch_size]
            results = [processor(item) for item in batch]
            yield results

            # Force cleanup if threshold reached
            if i % (self.gc_threshold * batch_size) == 0:
                gc.collect()

    def optimize_for_large_collection(self, size: int) -> dict[str, Any]:
        """Get optimization recommendations for a collection size."""
        recommendations = {
            "batch_size": 10,
            "use_streaming": False,
            "enable_pooling": False,
            "memory_limit_mb": 100,
        }

        if size < 100:
            recommendations["batch_size"] = 10
        elif size < 1000:
            recommendations["batch_size"] = 50
            recommendations["enable_pooling"] = True
        else:
            recommendations["batch_size"] = 100
            recommendations["use_streaming"] = True
            recommendations["enable_pooling"] = True
            recommendations["memory_limit_mb"] = 500 + (size // 100)

        return recommendations


class MemoryOptimizerTransformer(ASTTransformer):
    """AST transformer that optimizes memory usage."""

    def __init__(self, string_pool: dict[str, str], aggressive: bool = False) -> None:
        super().__init__()
        self.string_pool = string_pool
        self.aggressive = aggressive
        self.nodes_processed = 0

    def visit(self, node: ASTNode) -> ASTNode:
        """Visit a node and optimize its memory usage."""
        self.nodes_processed += 1
        return super().visit(node)

    def visit_string_literal(self, node: Any) -> Any:
        """Pool string literals to reduce memory usage."""
        if hasattr(node, "value") and isinstance(node.value, str):
            # Use string pooling
            pooled = self.string_pool.get(node.value)
            if pooled is None:
                self.string_pool[node.value] = node.value
                pooled = node.value
            else:
                # Reuse existing string
                node.value = pooled
        return node

    def visit_identifier(self, node: Any) -> Any:
        """Pool identifier names."""
        if hasattr(node, "name") and isinstance(node.name, str):
            pooled = self.string_pool.get(node.name)
            if pooled is None:
                self.string_pool[node.name] = node.name
                pooled = node.name
            else:
                node.name = pooled
        return node

    def visit_rule(self, node: Rule) -> Rule:
        """Optimize rule memory usage."""
        # Pool rule name
        if node.name:
            pooled = self.string_pool.get(node.name)
            if pooled is None:
                self.string_pool[node.name] = node.name
            else:
                node.name = pooled

        # Visit children
        if node.condition:
            node.condition = self.visit(node.condition)

        if node.strings:
            node.strings = [self.visit(s) for s in node.strings]

        if node.meta:
            node.meta = [self.visit(m) for m in node.meta]

        if node.tags:
            node.tags = [self.visit(t) for t in node.tags]

        # Clear unnecessary attributes if aggressive
        if self.aggressive and hasattr(node, "location"):
            # Remove location info if not needed
            node.location = None

        return node

    def visit_plain_string(self, node: Any) -> Any:
        """Optimize plain string memory usage."""
        # Pool string value
        if hasattr(node, "value") and isinstance(node.value, str):
            pooled = self.string_pool.get(node.value)
            if pooled is None:
                self.string_pool[node.value] = node.value
            else:
                node.value = pooled

        # Pool identifier
        if hasattr(node, "identifier") and isinstance(node.identifier, str):
            pooled = self.string_pool.get(node.identifier)
            if pooled is None:
                self.string_pool[node.identifier] = node.identifier
            else:
                node.identifier = pooled

        return node

    def visit_meta(self, node: Any) -> Any:
        """Optimize meta memory usage."""
        # Pool meta key
        if hasattr(node, "key") and isinstance(node.key, str):
            pooled = self.string_pool.get(node.key)
            if pooled is None:
                self.string_pool[node.key] = node.key
            else:
                node.key = pooled

        # Pool string values
        if hasattr(node, "value") and isinstance(node.value, str):
            pooled = self.string_pool.get(node.value)
            if pooled is None:
                self.string_pool[node.value] = node.value
            else:
                node.value = pooled

        return node

    def visit_tag(self, node: Any) -> Any:
        """Optimize tag memory usage."""
        if hasattr(node, "name") and isinstance(node.name, str):
            pooled = self.string_pool.get(node.name)
            if pooled is None:
                self.string_pool[node.name] = node.name
            else:
                node.name = pooled
        return node

    # Pass-through methods for other node types
    def visit_yara_file(self, node: YaraFile) -> YaraFile:
        """Optimize YaraFile memory usage."""
        if node.imports:
            node.imports = [self.visit(imp) for imp in node.imports]
        if node.includes:
            node.includes = [self.visit(inc) for inc in node.includes]
        if node.rules:
            node.rules = [self.visit(rule) for rule in node.rules]
        return node

    def visit_import(self, node: Any) -> Any:
        """Optimize import memory usage."""
        if hasattr(node, "module") and isinstance(node.module, str):
            pooled = self.string_pool.get(node.module)
            if pooled is None:
                self.string_pool[node.module] = node.module
            else:
                node.module = pooled
        return node

    def visit_include(self, node: Any) -> Any:
        """Optimize include memory usage."""
        if hasattr(node, "path") and isinstance(node.path, str):
            pooled = self.string_pool.get(node.path)
            if pooled is None:
                self.string_pool[node.path] = node.path
            else:
                node.path = pooled
        return node

    def visit_boolean_literal(self, node: Any) -> Any:
        return node

    def visit_integer_literal(self, node: Any) -> Any:
        return node

    def visit_double_literal(self, node: Any) -> Any:
        return node

    def visit_string_identifier(self, node: Any) -> Any:
        if hasattr(node, "name") and isinstance(node.name, str):
            pooled = self.string_pool.get(node.name)
            if pooled is None:
                self.string_pool[node.name] = node.name
            else:
                node.name = pooled
        return node

    def visit_string_wildcard(self, node: Any) -> Any:
        if hasattr(node, "pattern") and isinstance(node.pattern, str):
            pooled = self.string_pool.get(node.pattern)
            if pooled is None:
                self.string_pool[node.pattern] = node.pattern
            else:
                node.pattern = pooled
        return node

    def visit_binary_expression(self, node: Any) -> Any:
        if hasattr(node, "left"):
            node.left = self.visit(node.left)
        if hasattr(node, "right"):
            node.right = self.visit(node.right)
        if hasattr(node, "operator") and isinstance(node.operator, str):
            pooled = self.string_pool.get(node.operator)
            if pooled is None:
                self.string_pool[node.operator] = node.operator
            else:
                node.operator = pooled
        return node

    def visit_unary_expression(self, node: Any) -> Any:
        if hasattr(node, "operand"):
            node.operand = self.visit(node.operand)
        if hasattr(node, "operator") and isinstance(node.operator, str):
            pooled = self.string_pool.get(node.operator)
            if pooled is None:
                self.string_pool[node.operator] = node.operator
            else:
                node.operator = pooled
        return node

    def visit_hex_string(self, node: Any) -> Any:
        if hasattr(node, "identifier") and isinstance(node.identifier, str):
            pooled = self.string_pool.get(node.identifier)
            if pooled is None:
                self.string_pool[node.identifier] = node.identifier
            else:
                node.identifier = pooled
        return node

    def visit_regex_string(self, node: Any) -> Any:
        if hasattr(node, "identifier") and isinstance(node.identifier, str):
            pooled = self.string_pool.get(node.identifier)
            if pooled is None:
                self.string_pool[node.identifier] = node.identifier
            else:
                node.identifier = pooled
        if hasattr(node, "regex") and isinstance(node.regex, str):
            pooled = self.string_pool.get(node.regex)
            if pooled is None:
                self.string_pool[node.regex] = node.regex
            else:
                node.regex = pooled
        return node
